from typing import cast, Self, Sequence, Final, NamedTuple
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    JaxState,
    JaxStateSequence,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxBicycleObstacleStatesHistory,
    BicycleState,
    BicycleStateSequence,
    BicycleStateBatch,
    BicycleControlInputSequence,
    BicycleControlInputBatch,
    BicyclePositions,
    BicycleD_x,
    BICYCLE_D_X,
    BicycleD_u,
    BICYCLE_D_U,
    BicycleD_o,
    BICYCLE_D_O,
    POSE_D_O,
    DynamicalModel,
    ObstacleModel,
    ObstacleStateEstimator,
    EstimatedObstacleStates,
    JaxGaussianBelief,
    JaxNoiseCovarianceDescription,
    JaxNoiseModelProvider,
)
from faran.filters import (
    JaxExtendedKalmanFilter,
    JaxUnscentedKalmanFilter,
    jax_kalman_filter,
)
from faran.obstacles import JaxObstacle2dPoses
from faran.models.common import SMALL_UNCERTAINTY, LARGE_UNCERTAINTY
from faran.models.accelerated import invalid_obstacle_filter_from

from jaxtyping import Array as JaxArray, Float, Scalar
from numtypes import D


import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np


NO_LIMITS: Final = (jnp.asarray(-jnp.inf), jnp.asarray(jnp.inf))

BICYCLE_ESTIMATION_D_X: Final = 6
BICYCLE_OBSERVATION_D_O: Final = 3

type BicycleEstimationD_x = D[6]
type BicycleObservationD_o = D[3]

type StateArray = Float[JaxArray, f"{BICYCLE_D_X}"]
type ControlInputSequenceArray = Float[JaxArray, f"T {BICYCLE_D_U}"]
type StateBatchArray = Float[JaxArray, f"T {BICYCLE_D_X} M"]
type ControlInputBatchArray = Float[JaxArray, f"T {BICYCLE_D_U} M"]

type StatesAtTimeStep = Float[JaxArray, f"{BICYCLE_D_X} M"]
type ControlInputsAtTimeStep = Float[JaxArray, f"{BICYCLE_D_U} M"]

type EstimationStateCovarianceArray = Float[
    JaxArray, f"{BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X}"
]
type ProcessNoiseCovarianceArray = Float[
    JaxArray, f"{BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X}"
]
type ObservationNoiseCovarianceArray = Float[
    JaxArray, f"{BICYCLE_OBSERVATION_D_O} {BICYCLE_OBSERVATION_D_O}"
]
type ObservationMatrix = Float[
    JaxArray, f"{BICYCLE_OBSERVATION_D_O} {BICYCLE_ESTIMATION_D_X}"
]

type BicycleGaussianBelief = JaxGaussianBelief
type JaxBicycleObstacleCovariances = Float[
    JaxArray, f"{BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X} K"
]
type KalmanFilter = JaxExtendedKalmanFilter | JaxUnscentedKalmanFilter


@jaxtyped
@dataclass(frozen=True)
class JaxBicycleState(BicycleState, JaxState):
    """Kinematic bicycle state: [x, y, heading, speed]."""

    _array: StateArray

    @staticmethod
    def create(
        *,
        x: float | Scalar,
        y: float | Scalar,
        heading: float | Scalar,
        speed: float | Scalar,
    ) -> "JaxBicycleState":
        return JaxBicycleState(jnp.array([x, y, heading, speed]))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, " BicycleD_x"]:
        return np.asarray(self.array)

    @property
    def dimension(self) -> BicycleD_x:
        return cast(BicycleD_x, self.array.shape[0])

    @property
    def x(self) -> float:
        return float(self.array[0])

    @property
    def y(self) -> float:
        return float(self.array[1])

    @property
    def heading(self) -> float:
        return float(self.array[2])

    @property
    def speed(self) -> float:
        return float(self.array[3])

    @property
    def array(self) -> StateArray:
        return self._array

    @property
    def x_scalar(self) -> Scalar:
        return self.array[0]

    @property
    def y_scalar(self) -> Scalar:
        return self.array[1]

    @property
    def heading_scalar(self) -> Scalar:
        return self.array[2]

    @property
    def speed_scalar(self) -> Scalar:
        return self.array[3]


@dataclass(kw_only=True, frozen=True)
class JaxBicycleStateSequence(BicycleStateSequence, JaxStateSequence):
    batch: "JaxBicycleStateBatch"
    rollout: int

    @staticmethod
    def of_states(states: Sequence[JaxBicycleState]) -> "JaxBicycleStateSequence":
        assert len(states) > 0, "States sequence must not be empty."

        array = jnp.stack([state.array for state in states], axis=0)[:, :, jnp.newaxis]

        return JaxBicycleStateSequence(
            batch=JaxBicycleStateBatch.wrap(array), rollout=0
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T BicycleD_x"]:
        return np.asarray(self.array)

    def step(self, index: int) -> JaxBicycleState:
        return JaxBicycleState(self.array[index])

    def batched(self) -> "JaxBicycleStateBatch":
        return JaxBicycleStateBatch.wrap(self.array[..., jnp.newaxis])

    def x(self) -> Float[Array, " T"]:
        return np.asarray(self.array[:, 0])

    def y(self) -> Float[Array, " T"]:
        return np.asarray(self.array[:, 1])

    def heading(self) -> Float[Array, " T"]:
        return np.asarray(self.array[:, 2])

    def speed(self) -> Float[Array, " T"]:
        return np.asarray(self.array[:, 3])

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_x:
        return cast(BicycleD_x, self.array.shape[1])

    @property
    def array(self) -> Float[JaxArray, f"T {BICYCLE_D_X}"]:
        return self.batch.array[:, :, self.rollout]


@jaxtyped
@dataclass(frozen=True)
class JaxBicycleStateBatch(BicycleStateBatch, JaxStateBatch):
    _array: StateBatchArray

    @staticmethod
    def wrap(array: StateBatchArray) -> "JaxBicycleStateBatch":
        return JaxBicycleStateBatch(array)

    @staticmethod
    def of_states(states: Sequence[JaxBicycleState]) -> "JaxBicycleStateBatch":
        assert len(states) > 0, "States sequence must not be empty."

        array = jnp.stack([state.array for state in states], axis=0)[:, :, jnp.newaxis]

        return JaxBicycleStateBatch(array)

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T BicycleD_x M"]:
        return np.asarray(self.array)

    def heading(self) -> Float[Array, "T M"]:
        return np.asarray(self.array[:, 2, :])

    def speed(self) -> Float[Array, "T M"]:
        return np.asarray(self.array[:, 3, :])

    def rollout(self, index: int) -> JaxBicycleStateSequence:
        return JaxBicycleStateSequence(batch=self, rollout=index)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_x:
        return cast(BicycleD_x, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def positions(self) -> "JaxBicyclePositions":
        return JaxBicyclePositions(self)

    @property
    def array(self) -> StateBatchArray:
        return self._array

    @property
    def heading_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 2, :]


@jaxtyped
@dataclass(frozen=True)
class JaxBicyclePositions(BicyclePositions):
    batch: JaxBicycleStateBatch

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 M"]:
        return np.asarray(self.batch.array[:, :2, :])

    def x(self) -> Float[Array, "T M"]:
        return np.asarray(self.batch.array[:, 0, :])

    def y(self) -> Float[Array, "T M"]:
        return np.asarray(self.batch.array[:, 1, :])

    @property
    def x_array(self) -> Float[JaxArray, "T M"]:
        return self.batch.array[:, 0, :]

    @property
    def y_array(self) -> Float[JaxArray, "T M"]:
        return self.batch.array[:, 1, :]

    @property
    def array(self) -> Float[JaxArray, "T 2 M"]:
        return self.batch.array[:, :2, :]

    @property
    def horizon(self) -> int:
        return self.batch.horizon

    @property
    def rollout_count(self) -> int:
        return self.batch.rollout_count


@jaxtyped
@dataclass(frozen=True)
class JaxBicycleControlInputSequence(
    BicycleControlInputSequence, JaxControlInputSequence
):
    """Control inputs: [acceleration, steering_angle]."""

    _array: ControlInputSequenceArray

    @staticmethod
    def zeroes(horizon: int) -> "JaxBicycleControlInputSequence":
        return JaxBicycleControlInputSequence(jnp.zeros((horizon, BICYCLE_D_U)))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T BicycleD_u"]:
        return np.asarray(self.array)

    def similar(self, *, array: Float[JaxArray, "L D_u"]) -> Self:
        return self.__class__(array)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_u:
        return cast(BicycleD_u, self.array.shape[1])

    @property
    def array(self) -> ControlInputSequenceArray:
        return self._array

    @property
    def accelerations_array(self) -> Float[JaxArray, " T"]:
        return self.array[:, 0]

    @property
    def steering_angles_array(self) -> Float[JaxArray, " T"]:
        return self.array[:, 1]


@jaxtyped
@dataclass(frozen=True)
class JaxBicycleControlInputBatch(BicycleControlInputBatch, JaxControlInputBatch):
    _array: ControlInputBatchArray

    @staticmethod
    def zero(*, horizon: int, rollout_count: int = 1) -> "JaxBicycleControlInputBatch":
        array = jnp.zeros((horizon, BICYCLE_D_U, rollout_count))

        return JaxBicycleControlInputBatch(array)

    @staticmethod
    def create(
        *, array: Float[Array, "T BicycleD_u M"] | ControlInputBatchArray
    ) -> "JaxBicycleControlInputBatch":
        return JaxBicycleControlInputBatch(jnp.asarray(array))

    @staticmethod
    def of(sequence: JaxBicycleControlInputSequence) -> "JaxBicycleControlInputBatch":
        array = sequence.array[..., jnp.newaxis]

        assert array.shape == (expected := (sequence.horizon, BICYCLE_D_U, 1)), (
            f"Array shape {array.shape} does not match expected shape {expected}."
        )

        return JaxBicycleControlInputBatch(array)

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T BicycleD_u M"]:
        return np.asarray(self.array)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_u:
        return cast(BicycleD_u, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> ControlInputBatchArray:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class JaxBicycleObstacleStates:
    _array: Float[JaxArray, f"{BICYCLE_D_O} K"]

    @staticmethod
    def wrap(
        array: Float[Array, "BicycleD_o K"] | Float[JaxArray, f"{BICYCLE_D_O} K"],
    ) -> "JaxBicycleObstacleStates":
        return JaxBicycleObstacleStates(jnp.asarray(array))

    @staticmethod
    def create(
        *,
        x: Float[Array, " K"] | Float[JaxArray, " K"],
        y: Float[Array, " K"] | Float[JaxArray, " K"],
        heading: Float[Array, " K"] | Float[JaxArray, " K"],
        speed: Float[Array, " K"] | Float[JaxArray, " K"],
    ) -> "JaxBicycleObstacleStates":
        return JaxBicycleObstacleStates(jnp.stack([x, y, heading, speed], axis=0))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "BicycleD_o K"]:
        return self._numpy_array

    def x(self) -> Float[Array, " K"]:
        return self._numpy_array[0, :]

    def y(self) -> Float[Array, " K"]:
        return self._numpy_array[1, :]

    def heading(self) -> Float[Array, " K"]:
        return self._numpy_array[2, :]

    def speed(self) -> Float[Array, " K"]:
        return self._numpy_array[3, :]

    @property
    def dimension(self) -> BicycleD_o:
        return cast(BicycleD_o, self.array.shape[0])

    @property
    def count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[JaxArray, f"{BICYCLE_D_O} K"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "BicycleD_o K"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxBicycleObstacleStateSequences:
    _array: Float[JaxArray, f"T {BICYCLE_ESTIMATION_D_X} K"]
    _covariance: Float[
        JaxArray, f"T {BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X} K"
    ]

    @staticmethod
    def wrap(
        *,
        array: Float[JaxArray, f"T {BICYCLE_ESTIMATION_D_X} K"],
        covariance: Float[
            JaxArray, f"T {BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X} K"
        ],
    ) -> "JaxBicycleObstacleStateSequences":
        return JaxBicycleObstacleStateSequences(array, covariance)

    @staticmethod
    def create(
        predictions: Sequence[BicycleGaussianBelief],
    ) -> "JaxBicycleObstacleStateSequences":
        assert len(predictions) > 0, "Predictions sequence must not be empty."

        array = jnp.stack([belief.mean for belief in predictions], axis=0)
        covariance = jnp.stack([belief.covariance for belief in predictions], axis=0)

        return JaxBicycleObstacleStateSequences(array, covariance)

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T BicycleEstimationD_x K"]:
        return self._numpy_array

    def x(self) -> Float[Array, "T K"]:
        return self._numpy_array[:, 0, :]

    def y(self) -> Float[Array, "T K"]:
        return self._numpy_array[:, 1, :]

    def heading(self) -> Float[Array, "T K"]:
        return self._numpy_array[:, 2, :]

    def covariance(
        self,
    ) -> Float[JaxArray, f"T {BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X} K"]:
        return self._covariance

    def pose(self) -> JaxObstacle2dPoses:
        return JaxObstacle2dPoses.create(
            x=self.x_array,
            y=self.y_array,
            heading=self.heading_array,
            covariance=self.pose_covariance_array,
        )

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleEstimationD_x:
        return cast(BicycleEstimationD_x, self.array.shape[1])

    @property
    def count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[JaxArray, f"T {BICYCLE_ESTIMATION_D_X} K"]:
        return self._array

    @property
    def x_array(self) -> Float[JaxArray, "T K"]:
        return self._array[:, 0, :]

    @property
    def y_array(self) -> Float[JaxArray, "T K"]:
        return self._array[:, 1, :]

    @property
    def heading_array(self) -> Float[JaxArray, "T K"]:
        return self._array[:, 2, :]

    @property
    def covariance_array(
        self,
    ) -> Float[JaxArray, f"T {BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X} K"]:
        return self._covariance

    @property
    def pose_covariance_array(self) -> Float[JaxArray, f"T {POSE_D_O} {POSE_D_O} K"]:
        return self._covariance[:, :3, :3, :]

    @cached_property
    def _numpy_array(self) -> Float[Array, "T BicycleEstimationD_x K"]:
        return np.asarray(self._array)


@jaxtyped
@dataclass(frozen=True)
class JaxBicycleObstacleInputs:
    _accelerations: Float[JaxArray, " K"]
    _steering_angles: Float[JaxArray, " K"]

    @staticmethod
    def wrap(
        inputs: Float[Array, "BicycleD_u K"] | Float[JaxArray, f"{BICYCLE_D_U} K"],
    ) -> "JaxBicycleObstacleInputs":
        inputs = jnp.asarray(inputs)
        return JaxBicycleObstacleInputs(
            _accelerations=inputs[0], _steering_angles=inputs[1]
        )

    @staticmethod
    def create(
        *,
        accelerations: Float[Array, " K"] | Float[JaxArray, " K"],
        steering_angles: Float[Array, " K"] | Float[JaxArray, " K"],
    ) -> "JaxBicycleObstacleInputs":
        return JaxBicycleObstacleInputs(
            _accelerations=jnp.asarray(accelerations),
            _steering_angles=jnp.asarray(steering_angles),
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "BicycleD_u K"]:
        return self._numpy_array

    def accelerations(self) -> Float[Array, " K"]:
        return self._numpy_array[0, :]

    def steering_angles(self) -> Float[Array, " K"]:
        return self._numpy_array[1, :]

    def zeroed(
        self, *, acceleration: bool = False, steering_angle: bool = False
    ) -> "JaxBicycleObstacleInputs":
        """Returns a version of the inputs with the specified components zeroed out."""
        return JaxBicycleObstacleInputs(
            _accelerations=jnp.zeros_like(self._accelerations)
            if acceleration
            else self._accelerations,
            _steering_angles=jnp.zeros_like(self._steering_angles)
            if steering_angle
            else self._steering_angles,
        )

    @property
    def dimension(self) -> BicycleD_u:
        return BICYCLE_D_U

    @property
    def count(self) -> int:
        return self._steering_angles.shape[0]

    @property
    def accelerations_array(self) -> Float[JaxArray, " K"]:
        return self._accelerations

    @property
    def steering_angles_array(self) -> Float[JaxArray, " K"]:
        return self._steering_angles

    @property
    def array(self) -> Float[JaxArray, f"{BICYCLE_D_U} K"]:
        return jnp.stack([self._accelerations, self._steering_angles], axis=0)

    @cached_property
    def _numpy_array(self) -> Float[Array, "BicycleD_u K"]:
        return np.asarray(self.array)


@dataclass(kw_only=True, frozen=True)
class JaxBicycleModel(
    DynamicalModel[
        JaxBicycleState,
        JaxBicycleStateSequence,
        JaxBicycleStateBatch,
        JaxBicycleControlInputSequence,
        JaxBicycleControlInputBatch,
    ],
):
    """Kinematic bicycle with rear-axle reference, Euler-integrated with configurable limits."""

    _time_step_size: float
    time_step_size_scalar: Scalar
    wheelbase: Scalar
    rear_axle_distance: Scalar
    speed_limits: tuple[Scalar, Scalar]
    steering_limits: tuple[Scalar, Scalar]
    acceleration_limits: tuple[Scalar, Scalar]

    @staticmethod
    def create(
        *,
        time_step_size: float,
        wheelbase: float = 1.0,
        rear_axle_distance: float = 0.0,
        speed_limits: tuple[float, float] | None = None,
        steering_limits: tuple[float, float] | None = None,
        acceleration_limits: tuple[float, float] | None = None,
    ) -> "JaxBicycleModel":

        return JaxBicycleModel(
            _time_step_size=time_step_size,
            time_step_size_scalar=jnp.asarray(time_step_size),
            wheelbase=jnp.asarray(wheelbase),
            rear_axle_distance=jnp.asarray(rear_axle_distance),
            speed_limits=wrap(speed_limits) if speed_limits is not None else NO_LIMITS,
            steering_limits=wrap(steering_limits)
            if steering_limits is not None
            else NO_LIMITS,
            acceleration_limits=wrap(acceleration_limits)
            if acceleration_limits is not None
            else NO_LIMITS,
        )

    def simulate(
        self, inputs: JaxBicycleControlInputBatch, initial_state: JaxBicycleState
    ) -> JaxBicycleStateBatch:
        rollout_count = inputs.rollout_count

        initial = jnp.stack(
            [
                jnp.full(rollout_count, initial_state.x_scalar),
                jnp.full(rollout_count, initial_state.y_scalar),
                jnp.full(rollout_count, initial_state.heading_scalar),
                jnp.full(rollout_count, initial_state.speed_scalar),
            ]
        )

        return JaxBicycleStateBatch(
            simulate(
                inputs.array,
                initial,
                time_step_size=self.time_step_size_scalar,
                wheelbase=self.wheelbase,
                rear_axle_distance=self.rear_axle_distance,
                speed_limits=self.speed_limits,
                steering_limits=self.steering_limits,
                acceleration_limits=self.acceleration_limits,
            )
        )

    def step(
        self, inputs: JaxBicycleControlInputSequence, state: JaxBicycleState
    ) -> JaxBicycleState:
        return JaxBicycleState(
            step(
                state.array.reshape(-1, 1),
                inputs.array[0].reshape(-1, 1),
                time_step_size=self.time_step_size_scalar,
                wheelbase=self.wheelbase,
                rear_axle_distance=self.rear_axle_distance,
                speed_limits=self.speed_limits,
                steering_limits=self.steering_limits,
                acceleration_limits=self.acceleration_limits,
            )[:, 0]
        )

    def forward(
        self, inputs: JaxBicycleControlInputSequence, state: JaxBicycleState
    ) -> JaxBicycleStateSequence:
        return self.simulate(JaxBicycleControlInputBatch.of(inputs), state).rollout(0)

    @property
    def time_step_size(self) -> float:
        return self._time_step_size


class JaxBicycleStateEstimationModel(NamedTuple):
    """Kinematic bicycle model used for state estimation."""

    time_step_size: Scalar
    wheelbase: Scalar
    rear_axle_distance: Scalar
    initial_state_covariance: EstimationStateCovarianceArray

    @staticmethod
    def create(
        *,
        time_step_size: float,
        wheelbase: float,
        rear_axle_distance: float = 0.0,
        initial_state_covariance: Float[
            Array, "BicycleEstimationD_x BicycleEstimationD_x"
        ]
        | EstimationStateCovarianceArray
        | None = None,
    ) -> "JaxBicycleStateEstimationModel":
        if initial_state_covariance is None:
            initial_state_covariance = (
                JaxBicycleStateEstimationModel.default_initial_state_covariance()
            )

        return JaxBicycleStateEstimationModel(
            time_step_size=jnp.asarray(time_step_size),
            wheelbase=jnp.asarray(wheelbase),
            rear_axle_distance=jnp.asarray(rear_axle_distance),
            initial_state_covariance=jnp.asarray(initial_state_covariance),
        )

    @staticmethod
    def default_initial_state_covariance() -> EstimationStateCovarianceArray:
        # NOTE: Sure of pose, unsure of everything else.
        # Steering angle covariance is kept somewhat smaller (0.1), since it
        # otherwise messes up UKF.
        return jnp.diag(
            jnp.array(
                [
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                    0.1,
                ]
            )
        )

    @jax.jit
    @jaxtyped
    def __call__(self, state: Float[JaxArray, "D_x K"]) -> Float[JaxArray, "D_x K"]:
        return jax.vmap(self.step, in_axes=1, out_axes=1)(state)

    @jax.jit
    @jaxtyped
    def jacobian(self, state: Float[JaxArray, "D_x K"]) -> Float[JaxArray, "D_x D_x K"]:
        return jax.vmap(jax.jacfwd(self.step))(state.T).transpose(1, 2, 0)

    @jax.jit
    @jaxtyped
    def step(self, state: Float[JaxArray, " D_x"]) -> Float[JaxArray, " D_x"]:
        dt = self.time_step_size
        l_r = self.rear_axle_distance
        x, y, theta, v, a, delta = state

        beta = jnp.arctan(l_r / self.wheelbase * jnp.tan(delta))

        return jnp.array(
            [
                x + v * jnp.cos(theta + beta) * dt,
                y + v * jnp.sin(theta + beta) * dt,
                theta + v * jnp.cos(beta) * jnp.tan(delta) / self.wheelbase * dt,
                v + a * dt,
                a,
                delta,
            ]
        )

    def observations_from(
        self, history: JaxBicycleObstacleStatesHistory
    ) -> Float[JaxArray, "T D_z K"]:
        return jnp.stack(
            [history.x_array, history.y_array, history.heading_array], axis=1
        )

    def states_from(self, belief: BicycleGaussianBelief) -> JaxBicycleObstacleStates:
        return JaxBicycleObstacleStates.wrap(belief.mean[:4, :])

    def inputs_from(self, belief: BicycleGaussianBelief) -> JaxBicycleObstacleInputs:
        return JaxBicycleObstacleInputs.wrap(belief.mean[4:, :])

    def initial_belief_from(
        self,
        *,
        states: JaxBicycleObstacleStates,
        inputs: JaxBicycleObstacleInputs,
        covariances: JaxBicycleObstacleCovariances | None = None,
    ) -> BicycleGaussianBelief:
        augmented = jnp.concatenate([states.array, inputs.array], axis=0)

        if covariances is None:
            # NOTE: No covariance means we are "certain" about the states.
            covariances = jnp.broadcast_to(
                jnp.eye(BICYCLE_ESTIMATION_D_X)[:, :, jnp.newaxis] * SMALL_UNCERTAINTY,
                (BICYCLE_ESTIMATION_D_X, BICYCLE_ESTIMATION_D_X, states.count),
            )

        return JaxGaussianBelief(mean=augmented, covariance=covariances)

    @property
    def observation_matrix(self) -> ObservationMatrix:
        # NOTE: We observe the pose directly.
        return jnp.array(
            [
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
            ]
        )


@dataclass(kw_only=True, frozen=True)
class JaxBicycleObstacleModel(
    ObstacleModel[
        JaxBicycleObstacleStatesHistory,
        JaxBicycleObstacleStates,
        JaxBicycleObstacleInputs,
        JaxBicycleObstacleCovariances,
        JaxBicycleObstacleStateSequences,
    ]
):
    """Propagates bicycle kinematics forward given states and control inputs."""

    model: "JaxBicycleStateEstimationModel"
    process_noise_covariance: "EstimationStateCovarianceArray"
    predictor: KalmanFilter

    @staticmethod
    def unscented(
        *,
        time_step_size: float,
        wheelbase: float = 1.0,
        rear_axle_distance: float = 0.0,
        process_noise_covariance: JaxNoiseCovarianceDescription = 1e-3,
        sigma_point_spread: float = 1.0,
        prior_knowledge: float = 2.0,
    ) -> "JaxBicycleObstacleModel":
        """Creates a bicycle obstacle model for obstacle state prediction. This
        model uses the unscented transform for propagating state information through
        the nonlinear bicycle dynamics.

        Args:
            time_step_size: Time step size for state propagation.
            wheelbase: Distance between front and rear axles.
            rear_axle_distance: Distance from the reference point to the rear axle.
            process_noise_covariance: The process noise covariance, either as a
                full covariance array, a diagonal covariance vector, or a scalar
                variance representing isotropic noise.
            sigma_point_spread: Spread of sigma points (α) for the unscented
                transform, controlling how far the sigma points are from the mean.
            prior_knowledge: Prior knowledge (β) about the state distribution
                for the unscented transform.
        """
        standardized_covariance = jax_kalman_filter.standardize_noise_covariance(
            process_noise_covariance, dimension=BICYCLE_ESTIMATION_D_X
        )
        return JaxBicycleObstacleModel(
            model=JaxBicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                wheelbase=wheelbase,
                rear_axle_distance=rear_axle_distance,
            ),
            process_noise_covariance=standardized_covariance,
            predictor=JaxUnscentedKalmanFilter.create(
                alpha=sigma_point_spread, beta=prior_knowledge
            ),
        )

    def forward(
        self,
        *,
        states: JaxBicycleObstacleStates,
        inputs: JaxBicycleObstacleInputs,
        covariances: JaxBicycleObstacleCovariances | None,
        horizon: int,
    ) -> JaxBicycleObstacleStateSequences:
        means, covariance = forward(
            initial=self.model.initial_belief_from(
                states=states, inputs=inputs, covariances=covariances
            ),
            model=self.model,
            predictor=self.predictor,
            process_noise_covariance=self.process_noise_covariance,
            horizon=horizon,
        )

        return JaxBicycleObstacleStateSequences.wrap(array=means, covariance=covariance)


@dataclass(frozen=True)
class JaxFiniteDifferenceBicycleStateEstimator(
    ObstacleStateEstimator[
        JaxBicycleObstacleStatesHistory,
        JaxBicycleObstacleStates,
        JaxBicycleObstacleInputs,
    ]
):
    time_step_size: Scalar
    wheelbase: Scalar
    rear_axle_distance: Scalar

    @staticmethod
    def create(
        *, time_step_size: float, wheelbase: float, rear_axle_distance: float = 0.0
    ) -> "JaxFiniteDifferenceBicycleStateEstimator":
        return JaxFiniteDifferenceBicycleStateEstimator(
            time_step_size=jnp.asarray(time_step_size),
            wheelbase=jnp.asarray(wheelbase),
            rear_axle_distance=jnp.asarray(rear_axle_distance),
        )

    def estimate_from(
        self, history: JaxBicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        JaxBicycleObstacleStates, JaxBicycleObstacleInputs, None
    ]:
        """Estimates current states and inputs from pose history using finite differences. Zeros
        are assumed for states that cannot be estimated due to insufficient history length.

        Computes the following quantities from the kinematic bicycle model:

        **Speed** (requires T ≥ 2):
            Magnitude of displacement with sign from heading projection:
            $$v_t = \\text{sign}(v_t) \\cdot \\frac{\\sqrt{(x_t - x_{t-1})^2 + (y_t - y_{t-1})^2}}{\\Delta t}$$

        **Acceleration** (requires T ≥ 3):
            Change in speed over time:
            $$a_t = \\frac{v_t - v_{t-1}}{\\Delta t}$$

        **Steering angle** (requires T ≥ 2):
            Derived from the kinematic bicycle model, where $\\omega = \\frac{d\\theta}{dt}$ is the heading rate:
            $$\\delta_t = \\arctan\\left(\\frac{L \\cdot \\omega_t}{\\sqrt{v_t^2 - l_r^2 \\omega_t^2}}\\right)$$
            Zero when $|v_t| < \\epsilon$ or $v_t^2 \\leq l_r^2 \\omega_t^2$ (estimate becomes unreliable).

        Args:
            history: Obstacle pose history containing at least one entry.
        """
        assert history.horizon > 0, "History must contain at least one state."

        estimated = estimate_states(
            x_history=history.x_array,
            y_history=history.y_array,
            heading_history=history.heading_array,
            time_step_size=self.time_step_size,
            wheelbase=self.wheelbase,
            rear_axle_distance=self.rear_axle_distance,
        )

        return EstimatedObstacleStates(
            states=JaxBicycleObstacleStates.create(
                x=estimated.x,
                y=estimated.y,
                heading=estimated.heading,
                speed=estimated.speed,
            ),
            inputs=JaxBicycleObstacleInputs.create(
                accelerations=estimated.accelerations,
                steering_angles=estimated.steering_angles,
            ),
            covariance=None,
        )


@dataclass(frozen=True)
class JaxKfBicycleStateEstimator(
    ObstacleStateEstimator[
        JaxBicycleObstacleStatesHistory,
        JaxBicycleObstacleStates,
        JaxBicycleObstacleInputs,
        JaxBicycleObstacleCovariances,
    ]
):
    """Kalman Filter state estimator for bicycle model obstacles."""

    process_noise_covariance: ProcessNoiseCovarianceArray
    observation_noise_covariance: ObservationNoiseCovarianceArray
    model: JaxBicycleStateEstimationModel
    estimator: KalmanFilter
    noise_model: JaxNoiseModelProvider | None = None

    @staticmethod
    def ekf(
        *,
        time_step_size: float,
        wheelbase: float,
        rear_axle_distance: float = 0.0,
        process_noise_covariance: JaxNoiseCovarianceDescription,
        observation_noise_covariance: JaxNoiseCovarianceDescription,
        initial_state_covariance: Float[
            Array, "BicycleEstimationD_x BicycleEstimationD_x"
        ]
        | EstimationStateCovarianceArray
        | None = None,
        noise_model: JaxNoiseModelProvider | None = None,
    ) -> "JaxKfBicycleStateEstimator":
        """Creates an EKF state estimator for the bicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            wheelbase: The wheelbase of the bicycle.
            rear_axle_distance: Distance from the reference point to the rear axle.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for speed and inputs.
            noise_model: Optional noise model provider for adaptive noise estimation.
        """
        return JaxKfBicycleStateEstimator(
            process_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=BICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=BICYCLE_OBSERVATION_D_O
            ),
            model=JaxBicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                wheelbase=wheelbase,
                rear_axle_distance=rear_axle_distance,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=JaxExtendedKalmanFilter.create(noise_model=noise_model),
        )

    @staticmethod
    def ukf(
        *,
        time_step_size: float,
        wheelbase: float,
        rear_axle_distance: float = 0.0,
        process_noise_covariance: JaxNoiseCovarianceDescription,
        observation_noise_covariance: JaxNoiseCovarianceDescription,
        initial_state_covariance: Float[
            Array, "BicycleEstimationD_x BicycleEstimationD_x"
        ]
        | EstimationStateCovarianceArray
        | None = None,
        noise_model: JaxNoiseModelProvider | None = None,
    ) -> "JaxKfBicycleStateEstimator":
        """Creates a UKF state estimator for the bicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            wheelbase: The wheelbase of the bicycle.
            rear_axle_distance: Distance from the reference point to the rear axle.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for speed and inputs.
            noise_model: Optional noise model provider for adaptive noise estimation.
        """
        return JaxKfBicycleStateEstimator(
            process_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=BICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=BICYCLE_OBSERVATION_D_O
            ),
            model=JaxBicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                wheelbase=wheelbase,
                rear_axle_distance=rear_axle_distance,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=JaxUnscentedKalmanFilter.create(noise_model=noise_model),
        )

    def estimate_from(
        self, history: JaxBicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        JaxBicycleObstacleStates,
        JaxBicycleObstacleInputs,
        JaxBicycleObstacleCovariances,
    ]:
        estimate = self.estimator.filter(
            self.model.observations_from(history),
            initial_state_covariance=self.model.initial_state_covariance,
            state_transition=self.model,
            process_noise_covariance=self.process_noise_covariance,
            observation_noise_covariance=self.observation_noise_covariance,
            observation_matrix=self.model.observation_matrix,
        )

        return EstimatedObstacleStates(
            states=self.model.states_from(estimate),
            inputs=self.model.inputs_from(estimate),
            covariance=estimate.covariance,
        )


class EstimatedBicycleObstacleStates(NamedTuple):
    x: Float[JaxArray, " K"]
    y: Float[JaxArray, " K"]
    heading: Float[JaxArray, " K"]
    speed: Float[JaxArray, " K"]
    accelerations: Float[JaxArray, " K"]
    steering_angles: Float[JaxArray, " K"]


def wrap(limits: tuple[float, float]) -> tuple[Scalar, Scalar]:
    return (jnp.asarray(limits[0]), jnp.asarray(limits[1]))


@jax.jit
@jaxtyped
def simulate(
    inputs: ControlInputBatchArray,
    initial: StatesAtTimeStep,
    *,
    time_step_size: Scalar,
    wheelbase: Scalar,
    rear_axle_distance: Scalar,
    speed_limits: tuple[Scalar, Scalar],
    steering_limits: tuple[Scalar, Scalar],
    acceleration_limits: tuple[Scalar, Scalar],
) -> StateBatchArray:
    def do_step(
        state: StatesAtTimeStep, control: ControlInputsAtTimeStep
    ) -> tuple[StatesAtTimeStep, StatesAtTimeStep]:
        new_state = step(
            state,
            control,
            time_step_size=time_step_size,
            wheelbase=wheelbase,
            rear_axle_distance=rear_axle_distance,
            speed_limits=speed_limits,
            steering_limits=steering_limits,
            acceleration_limits=acceleration_limits,
        )
        return new_state, new_state

    _, states = jax.lax.scan(do_step, initial, inputs)
    return states


@jax.jit
@jaxtyped
def step(
    state: StatesAtTimeStep,
    control: ControlInputsAtTimeStep,
    *,
    time_step_size: Scalar,
    wheelbase: Scalar,
    rear_axle_distance: Scalar,
    speed_limits: tuple[Scalar, Scalar],
    steering_limits: tuple[Scalar, Scalar],
    acceleration_limits: tuple[Scalar, Scalar],
) -> StatesAtTimeStep:
    x, y, theta, v = state[0], state[1], state[2], state[3]
    a, delta = control[0], control[1]
    acceleration = jnp.clip(a, *acceleration_limits)
    steering = jnp.clip(delta, *steering_limits)

    beta = jnp.arctan(rear_axle_distance / wheelbase * jnp.tan(steering))

    new_x = x + v * jnp.cos(theta + beta) * time_step_size
    new_y = y + v * jnp.sin(theta + beta) * time_step_size
    new_theta = (
        theta + v * jnp.cos(beta) * jnp.tan(steering) / wheelbase * time_step_size
    )
    new_v = jnp.clip(v + acceleration * time_step_size, *speed_limits)

    return jnp.stack([new_x, new_y, new_theta, new_v])


@jax.jit
@jaxtyped
def estimate_states(
    *,
    x_history: Float[JaxArray, "T K"],
    y_history: Float[JaxArray, "T K"],
    heading_history: Float[JaxArray, "T K"],
    time_step_size: Scalar,
    wheelbase: Scalar,
    rear_axle_distance: Scalar,
) -> "EstimatedBicycleObstacleStates":
    filter_invalid_short = invalid_obstacle_filter_from(
        x_history, y_history, heading_history, check_recent=2
    )
    filter_invalid_long = invalid_obstacle_filter_from(
        x_history, y_history, heading_history, check_recent=3
    )

    return EstimatedBicycleObstacleStates(
        x=x_history[-1],
        y=y_history[-1],
        heading=heading_history[-1],
        speed=filter_invalid_short(
            speed := estimate_speed(
                x_history=x_history,
                y_history=y_history,
                heading_history=heading_history,
                time_step_size=time_step_size,
            )
        ),
        accelerations=filter_invalid_long(
            estimate_acceleration(
                x_history=x_history,
                y_history=y_history,
                heading_history=heading_history,
                speed_current=speed,
                time_step_size=time_step_size,
            )
        ),
        steering_angles=filter_invalid_short(
            estimate_steering_angle(
                heading_history=heading_history,
                speed_current=speed,
                time_step_size=time_step_size,
                wheelbase=wheelbase,
                rear_axle_distance=rear_axle_distance,
            )
        ),
    )


@jax.jit
@jaxtyped
def estimate_speed(
    *,
    x_history: Float[JaxArray, "T K"],
    y_history: Float[JaxArray, "T K"],
    heading_history: Float[JaxArray, "T K"],
    time_step_size: Scalar,
) -> Float[JaxArray, " K"]:

    horizon = x_history.shape[0]
    obstacle_count = x_history.shape[1]

    def estimate(
        *,
        x_current: Float[JaxArray, " K"],
        y_current: Float[JaxArray, " K"],
        x_previous: Float[JaxArray, " K"],
        y_previous: Float[JaxArray, " K"],
        heading: Float[JaxArray, " K"],
    ) -> Float[JaxArray, " K"]:
        delta_x = x_current - x_previous
        delta_y = y_current - y_previous

        projection = delta_x * jnp.cos(heading) + delta_y * jnp.sin(heading)
        magnitude = jnp.sqrt(delta_x**2 + delta_y**2)
        speeds = jnp.sign(projection) * magnitude / time_step_size

        return speeds

    if horizon < 2:
        return jnp.zeros(obstacle_count)

    return estimate(
        x_current=x_history[-1],
        y_current=y_history[-1],
        x_previous=x_history[-2],
        y_previous=y_history[-2],
        heading=heading_history[-1],
    )


@jax.jit
@jaxtyped
def estimate_acceleration(
    *,
    x_history: Float[JaxArray, "T K"],
    y_history: Float[JaxArray, "T K"],
    heading_history: Float[JaxArray, "T K"],
    speed_current: Float[JaxArray, " K"],
    time_step_size: Scalar,
) -> Float[JaxArray, " K"]:
    horizon = x_history.shape[0]
    obstacle_count = x_history.shape[1]

    def estimate(
        *, speed_current: Float[JaxArray, " K"], speed_previous: Float[JaxArray, " K"]
    ) -> Float[JaxArray, " K"]:
        return (speed_current - speed_previous) / time_step_size

    if horizon < 3:
        return jnp.zeros(obstacle_count)

    return estimate(
        speed_current=speed_current,
        speed_previous=estimate_speed(
            x_history=x_history[:-1],
            y_history=y_history[:-1],
            heading_history=heading_history[:-1],
            time_step_size=time_step_size,
        ),
    )


@jax.jit
@jaxtyped
def estimate_steering_angle(
    *,
    heading_history: Float[JaxArray, "T K"],
    speed_current: Float[JaxArray, " K"],
    time_step_size: Scalar,
    wheelbase: Scalar,
    rear_axle_distance: Scalar,
) -> Float[JaxArray, " K"]:
    horizon = heading_history.shape[0]
    obstacle_count = heading_history.shape[1]

    def estimate(
        *,
        speed_current: Float[JaxArray, " K"],
        heading_current: Float[JaxArray, " K"],
        heading_previous: Float[JaxArray, " K"],
    ) -> Float[JaxArray, " K"]:
        heading_velocity = (heading_current - heading_previous) / time_step_size

        denominator_squared = (
            speed_current**2 - rear_axle_distance**2 * heading_velocity**2
        )
        is_reliable = (jnp.abs(speed_current) > 1e-6) & (denominator_squared > 0)

        steering_angles = jnp.where(
            is_reliable,
            jnp.arctan(
                heading_velocity
                * wheelbase
                * jnp.sign(speed_current)
                / jnp.sqrt(jnp.maximum(denominator_squared, 1e-12))
            ),
            jnp.zeros_like(speed_current),
        )

        return steering_angles

    if horizon < 2:
        return jnp.zeros(obstacle_count)

    return estimate(
        speed_current=speed_current,
        heading_current=heading_history[-1],
        heading_previous=heading_history[-2],
    )


@eqx.filter_jit
@jaxtyped
def forward(
    *,
    initial: BicycleGaussianBelief,
    model: JaxBicycleStateEstimationModel,
    predictor: KalmanFilter,
    process_noise_covariance: ProcessNoiseCovarianceArray,
    horizon: int,
) -> tuple[
    Float[JaxArray, f"T {BICYCLE_ESTIMATION_D_X} K"],
    Float[JaxArray, f"T {BICYCLE_ESTIMATION_D_X} {BICYCLE_ESTIMATION_D_X} K"],
]:
    def step(
        belief: BicycleGaussianBelief, _
    ) -> tuple[BicycleGaussianBelief, BicycleGaussianBelief]:
        next_belief = predictor.predict(
            belief=belief,
            state_transition=model,
            process_noise_covariance=process_noise_covariance,
        )

        return next_belief, next_belief

    _, beliefs = jax.lax.scan(step, initial, xs=None, length=horizon)

    return beliefs.mean, beliefs.covariance



from typing import Final

from faran.types import (
    ControlInputBatch,
    StateBatch,
    CostFunction,
    NumPyCosts,
    JaxCosts,
)
from faran.costs.basic import (
    NumPyContouringCost,
    NumPyLagCost,
    NumPyProgressCost,
    NumPyControlSmoothingCost,
    NumPyControlEffortCost,
)
from faran.costs.accelerated import (
    JaxContouringCost,
    JaxLagCost,
    JaxProgressCost,
    JaxControlSmoothingCost,
    JaxControlEffortCost,
)
from faran.costs.combined import CombinedCost, NumPyCostSumFunction, JaxCostSumFunction
from faran.costs.collision import NumPyCollisionCost, JaxCollisionCost
from faran.costs.distance import (
    NumPyCircleDistanceExtractor,
    JaxCircleDistanceExtractor,
    NumPySatDistanceExtractor,
    JaxSatDistanceExtractor,
)
from faran.costs.boundary import (
    NumPyBoundaryCost,
    NumPyFixedWidthBoundary,
    NumPyPiecewiseFixedWidthBoundary,
    JaxBoundaryCost,
    JaxFixedWidthBoundary,
    JaxPiecewiseFixedWidthBoundary,
)


class costs:
    class numpy:
        @staticmethod
        def combined[
            ControlInputBatchT: ControlInputBatch,
            StateBatchT: StateBatch,
            CostsT: NumPyCosts,
        ](
            *costs: CostFunction[ControlInputBatchT, StateBatchT, CostsT],
        ) -> CombinedCost[ControlInputBatchT, StateBatchT, CostsT]:
            """Creates a NumPy cost function combining all given cost functions by summation."""
            return CombinedCost(costs=list(costs), sum=NumPyCostSumFunction())

        class tracking:
            contouring: Final = NumPyContouringCost.create
            lag: Final = NumPyLagCost.create
            progress: Final = NumPyProgressCost.create

        class comfort:
            control_smoothing: Final = NumPyControlSmoothingCost.create
            control_effort: Final = NumPyControlEffortCost.create

        class safety:
            collision: Final = NumPyCollisionCost.create
            boundary: Final = NumPyBoundaryCost.create

    class jax:
        @staticmethod
        def combined[
            ControlInputBatchT: ControlInputBatch,
            StateBatchT: StateBatch,
            CostsT: JaxCosts,
        ](
            *costs: CostFunction[ControlInputBatchT, StateBatchT, CostsT],
        ) -> CombinedCost[ControlInputBatchT, StateBatchT, CostsT]:
            """Creates a JAX cost function combining all given cost functions by summation."""
            return CombinedCost(costs=list(costs), sum=JaxCostSumFunction())

        class tracking:
            contouring: Final = JaxContouringCost.create
            lag: Final = JaxLagCost.create
            progress: Final = JaxProgressCost.create

        class comfort:
            control_smoothing: Final = JaxControlSmoothingCost.create
            control_effort: Final = JaxControlEffortCost.create

        class safety:
            collision: Final = JaxCollisionCost.create
            boundary: Final = JaxBoundaryCost.create


class distance:
    class numpy:
        circles: Final = NumPyCircleDistanceExtractor.create
        sat: Final = NumPySatDistanceExtractor.create

    class jax:
        circles: Final = JaxCircleDistanceExtractor.create
        sat: Final = JaxSatDistanceExtractor.create


class boundary:
    class numpy:
        fixed_width: Final = NumPyFixedWidthBoundary.create
        piecewise_fixed_width: Final = NumPyPiecewiseFixedWidthBoundary.create

    class jax:
        fixed_width: Final = JaxFixedWidthBoundary.create
        piecewise_fixed_width: Final = JaxPiecewiseFixedWidthBoundary.create



from typing import Final, Sequence
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    DynamicalModel,
    ObstacleModel,
    ObstacleStateEstimator,
    NumPyIntegratorState,
    NumPyIntegratorStateSequence,
    NumPyIntegratorStateBatch,
    NumPyIntegratorControlInputSequence,
    NumPyIntegratorControlInputBatch,
    NumPyIntegratorObstacleStatesHistory,
    EstimatedObstacleStates,
    NumPyGaussianBelief,
    NumPyNoiseCovarianceArrayDescription,
    NumPyNoiseCovarianceDescription,
    NumPyNoiseModelProvider,
)
from faran.states import (
    NumPySimpleState as SimpleState,
    NumPySimpleStateSequence as SimpleStateSequence,
    NumPySimpleStateBatch as SimpleStateBatch,
    NumPySimpleControlInputBatch as SimpleControlInputBatch,
)
from faran.filters import NumPyKalmanFilter, numpy_kalman_filter
from faran.models.common import SMALL_UNCERTAINTY, LARGE_UNCERTAINTY
from faran.models.basic import invalid_obstacle_filter_from
from faran.models.integrator.common import (
    observation_dimension_from,
    kf_state_dimension_for,
)


from jaxtyping import Float

import numpy as np


NO_LIMITS: Final = (float("-inf"), float("inf"))

type NumPyIntegratorObstacleCovariances = Float[Array, "D_x D_x K"]


@jaxtyped
@dataclass(frozen=True)
class NumPyIntegratorObstacleStates:
    """Arbitrary obstacle states with no semantic meaning attached."""

    _array: Float[Array, "D_o K"]

    @staticmethod
    def create(*, array: Float[Array, "D_o K"]) -> "NumPyIntegratorObstacleStates":
        return NumPyIntegratorObstacleStates(_array=array)

    def __array__(self, dtype: None | type = None) -> Float[Array, "D_o K"]:
        return self.array

    @property
    def dimension(self) -> int:
        return self.array.shape[0]

    @property
    def count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[Array, "D_o K"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyIntegratorObstacleStateSequences:
    """Time-indexed obstacle state sequences for integrator model obstacles."""

    _array: Float[Array, "T D_x K"]
    _covariance: Float[Array, "T D_x D_x K"]

    @staticmethod
    def create(
        predictions: Sequence[NumPyGaussianBelief],
    ) -> "NumPyIntegratorObstacleStateSequences":
        assert len(predictions) > 0, "Predictions sequence must not be empty."

        return NumPyIntegratorObstacleStateSequences(
            _array=np.stack([belief.mean for belief in predictions], axis=0),
            _covariance=np.stack([belief.covariance for belief in predictions], axis=0),
        )

    def __array__(self, dtype: None | type = None) -> Float[Array, "T D_x K"]:
        return self.array

    def covariance(self) -> Float[Array, "T D_x D_x K"]:
        return self._covariance

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> int:
        return self.array.shape[1]

    @property
    def count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[Array, "T D_x K"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyIntegratorObstacleInputs:
    _array: Float[Array, "D_o K"]

    @staticmethod
    def create(*, array: Float[Array, "D_o K"]) -> "NumPyIntegratorObstacleInputs":
        return NumPyIntegratorObstacleInputs(_array=array)

    def __array__(self, dtype: None | type = None) -> Float[Array, "D_o K"]:
        return self.array

    def zeroed(self, *, at: tuple[int, ...]) -> "NumPyIntegratorObstacleInputs":
        """Returns new obstacle inputs with inputs at specified state dimensions zeroed out."""

        zeroed_array = self.array.copy()
        zeroed_array[at, :] = 0.0

        return NumPyIntegratorObstacleInputs.create(array=zeroed_array)

    @property
    def dimension(self) -> int:
        return self.array.shape[0]

    @property
    def count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[Array, "D_o K"]:
        return self._array


@dataclass(kw_only=True, frozen=True)
class NumPyIntegratorModel(
    DynamicalModel[
        NumPyIntegratorState,
        NumPyIntegratorStateSequence,
        NumPyIntegratorStateBatch,
        NumPyIntegratorControlInputSequence,
        NumPyIntegratorControlInputBatch,
    ]
):
    """Point-mass model with direct position control, used for obstacle prediction."""

    _time_step_size: float
    state_limits: tuple[float, float]
    velocity_limits: tuple[float, float]
    periodic: bool

    @staticmethod
    def create(
        *,
        time_step_size: float,
        state_limits: tuple[float, float] | None = None,
        velocity_limits: tuple[float, float] | None = None,
        periodic: bool = False,
    ) -> "NumPyIntegratorModel":
        """A NumPy integrator model where state = cumulative sum of controls.

        This model represents a particle that moves according to velocity commands.
        State dimension must equal control dimension (D_x == D_u).

        $$x_{t+1} = \\text{clip}(x_t + \\text{clip}(u_t,\\; v_{\\text{lim}}) \\cdot \\Delta t,\\; s_{\\text{lim}})$$

        Args:
            time_step_size: The time step size for the integrator.
            state_limits: Optional tuple of (min, max) limits for the state values.
            velocity_limits: Optional tuple of (min, max) limits for the velocity inputs.
            periodic: Whether to apply periodic boundary conditions based on state_limits.
        """
        return NumPyIntegratorModel(
            _time_step_size=time_step_size,
            state_limits=state_limits if state_limits is not None else NO_LIMITS,
            velocity_limits=velocity_limits
            if velocity_limits is not None
            else NO_LIMITS,
            periodic=periodic,
        )

    def __post_init__(self) -> None:
        if self.periodic:
            validate_periodic_state_limits(self.state_limits)

    def simulate(
        self,
        inputs: NumPyIntegratorControlInputBatch,
        initial_state: NumPyIntegratorState,
    ) -> SimpleStateBatch:
        clipped_inputs = np.clip(inputs.array, *self.velocity_limits)

        return SimpleStateBatch(
            wrap_periodic_batch(
                states=simulate(
                    inputs=clipped_inputs,
                    initial_states=initial_state.array[:, np.newaxis],
                    time_step=self.time_step_size,
                ),
                state_limits=self.state_limits,
            )
            if self.periodic
            else simulate_with_state_limits(
                inputs=clipped_inputs,
                initial_state=initial_state.array,
                time_step=self.time_step_size,
                state_limits=self.state_limits,
            )
            if self.has_state_limits
            else simulate(
                inputs=clipped_inputs,
                initial_states=initial_state.array[:, np.newaxis],
                time_step=self.time_step_size,
            )
        )

    def step(
        self,
        inputs: NumPyIntegratorControlInputSequence,
        state: NumPyIntegratorState,
    ) -> SimpleState:
        clipped_control = np.clip(inputs.array[0], *self.velocity_limits)
        unbounded = state.array + clipped_control * self.time_step_size
        new_state = (
            wrap_periodic_state(state=unbounded, state_limits=self.state_limits)
            if self.periodic
            else np.clip(unbounded, *self.state_limits)
        )

        return SimpleState(new_state)

    def forward(
        self,
        inputs: NumPyIntegratorControlInputSequence,
        state: NumPyIntegratorState,
    ) -> SimpleStateSequence:
        return self.simulate(
            inputs=SimpleControlInputBatch.of(inputs), initial_state=state
        ).rollout(0)

    @property
    def time_step_size(self) -> float:
        return self._time_step_size

    @property
    def has_state_limits(self) -> bool:
        return self.state_limits != NO_LIMITS


@dataclass(frozen=True)
class NumPyIntegratorStateEstimationModel:
    """Single integrator model used for obstacle state estimation."""

    time_step_size: float
    observation_dimension: int
    initial_state_covariance: Float[Array, "D_x D_x"]

    @staticmethod
    def create(
        *,
        time_step_size: float,
        observation_dimension: int,
        initial_state_covariance: Float[Array, "D_x D_x"] | None = None,
    ) -> "NumPyIntegratorStateEstimationModel":
        if initial_state_covariance is None:
            initial_state_covariance = NumPyIntegratorStateEstimationModel.default_initial_state_covariance_for(
                observation_dimension
            )

        return NumPyIntegratorStateEstimationModel(
            time_step_size=time_step_size,
            observation_dimension=observation_dimension,
            initial_state_covariance=initial_state_covariance,
        )

    @staticmethod
    def default_initial_state_covariance_for(
        observation_dimension: int,
    ) -> Float[Array, "D_x D_x"]:
        D_o_ = observation_dimension

        # NOTE: We are sure of the observed states, unsure of the velocities.
        return np.diag(
            np.concatenate(
                (np.full(D_o_, SMALL_UNCERTAINTY), np.full(D_o_, LARGE_UNCERTAINTY))
            )
        )

    def __call__(self, augmented_state: Float[Array, "D_x K"]) -> Float[Array, "D_x K"]:
        return self.state_transition_matrix @ augmented_state

    def states_from(self, belief: NumPyGaussianBelief) -> NumPyIntegratorObstacleStates:
        D_o = self.observation_dimension
        return NumPyIntegratorObstacleStates.create(array=belief.mean[:D_o, :])

    def inputs_from(self, belief: NumPyGaussianBelief) -> NumPyIntegratorObstacleInputs:
        D_o = self.observation_dimension
        return NumPyIntegratorObstacleInputs.create(array=belief.mean[D_o:, :])

    def initial_belief_from(
        self,
        *,
        states: NumPyIntegratorObstacleStates,
        inputs: NumPyIntegratorObstacleInputs,
        covariances: NumPyIntegratorObstacleCovariances | None = None,
    ) -> NumPyGaussianBelief:
        augmented = np.concatenate([states.array, inputs.array], axis=0)
        D_x = 2 * self.observation_dimension

        if covariances is None:
            # NOTE: No covariance means we are "certain" about the states.
            covariances = np.broadcast_to(  # type: ignore
                np.eye(D_x)[:, :, np.newaxis] * SMALL_UNCERTAINTY,
                (D_x, D_x, states.count),
            ).copy()

        return NumPyGaussianBelief(mean=augmented, covariance=covariances)

    @cached_property
    def state_transition_matrix(self) -> Float[Array, "D_x D_x"]:
        D_o = self.observation_dimension

        # NOTE: State transition matrix for constant velocity model.
        return np.block(
            [
                [np.eye(D_o), self.time_step_size * np.eye(D_o)],
                [np.zeros((D_o, D_o)), np.eye(D_o)],
            ]
        )

    @cached_property
    def observation_matrix(self) -> Float[Array, "D_o D_x"]:
        D_o = self.observation_dimension

        # NOTE: We have a velocity for each observed state.
        return np.hstack((np.eye(D_o), np.zeros((D_o, D_o))))


@dataclass(kw_only=True, frozen=True)
class NumPyIntegratorObstacleModel(
    ObstacleModel[
        NumPyIntegratorObstacleStatesHistory,
        NumPyIntegratorObstacleStates,
        NumPyIntegratorObstacleInputs,
        NumPyIntegratorObstacleCovariances,
        NumPyIntegratorObstacleStateSequences,
    ]
):
    """Propagates integrator dynamics forward with constant velocity."""

    model: NumPyIntegratorStateEstimationModel
    process_noise_covariance: Float[Array, "D_x D_x"]
    predictor: NumPyKalmanFilter

    @staticmethod
    def create(
        *,
        time_step_size: float,
        state_dimension: int,
        process_noise_covariance: NumPyNoiseCovarianceDescription = 1e-3,
    ) -> "NumPyIntegratorObstacleModel":
        """Creates a NumPy integrator obstacle model.

        Args:
            time_step_size: The time step size for integration.
            state_dimension: The dimension of the obstacle state (e.g., 2 for 2D position).
            process_noise_covariance: The process noise covariance, either as a
                full covariance array, a diagonal covariance vector, or a scalar
                variance representing isotropic noise.
        """
        return NumPyIntegratorObstacleModel(
            model=NumPyIntegratorStateEstimationModel.create(
                time_step_size=time_step_size, observation_dimension=state_dimension
            ),
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance,
                dimension=kf_state_dimension_for(observation_dimension=state_dimension),
            ),
            predictor=NumPyKalmanFilter.create(),
        )

    def forward(
        self,
        *,
        states: NumPyIntegratorObstacleStates,
        inputs: NumPyIntegratorObstacleInputs,
        covariances: NumPyIntegratorObstacleCovariances | None,
        horizon: int,
    ) -> NumPyIntegratorObstacleStateSequences:
        beliefs = []
        last = self.model.initial_belief_from(
            states=states, inputs=inputs, covariances=covariances
        )

        for _ in range(horizon):
            beliefs.append(
                last := self.predictor.predict(
                    belief=last,
                    state_transition_matrix=self.model.state_transition_matrix,
                    process_noise_covariance=self.process_noise_covariance,
                )
            )

        return NumPyIntegratorObstacleStateSequences.create(beliefs)


@dataclass(frozen=True)
class NumPyFiniteDifferenceIntegratorStateEstimator(
    ObstacleStateEstimator[
        NumPyIntegratorObstacleStatesHistory,
        NumPyIntegratorObstacleStates,
        NumPyIntegratorObstacleInputs,
    ]
):
    time_step_size: float

    @staticmethod
    def create(
        *, time_step_size: float
    ) -> "NumPyFiniteDifferenceIntegratorStateEstimator":
        return NumPyFiniteDifferenceIntegratorStateEstimator(
            time_step_size=time_step_size
        )

    def estimate_from(
        self, history: NumPyIntegratorObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        NumPyIntegratorObstacleStates,
        NumPyIntegratorObstacleInputs,
        None,
    ]:
        """Estimates velocities from position history using finite differences.

        **Velocity** (requires T ≥ 2):
            $$v_t = (x_t - x_{t-1}) / (\\Delta t)$$
        """
        assert history.horizon > 0, (
            "History must contain at least one state for estimation."
        )

        filter_invalid = invalid_obstacle_filter_from(history, check_recent=2)
        velocities = self.estimate_velocities_from(history)

        return EstimatedObstacleStates(
            states=NumPyIntegratorObstacleStates.create(array=history.array[-1, :, :]),
            inputs=NumPyIntegratorObstacleInputs.create(
                array=filter_invalid(velocities)
            ),
            covariance=None,
        )

    def estimate_velocities_from(
        self, history: NumPyIntegratorObstacleStatesHistory
    ) -> Float[Array, "D_o K"]:
        if history.horizon < 2:
            return np.zeros((history.dimension, history.count))

        return self._estimate_velocities_from(
            current=history.array[-1, :, :],
            previous=history.array[-2, :, :],
        )

    def _estimate_velocities_from(
        self, *, current: Float[Array, "D_o K"], previous: Float[Array, "D_o K"]
    ) -> Float[Array, "D_o K"]:
        velocities = (current - previous) / self.time_step_size

        return velocities


@dataclass(frozen=True)
class NumPyKfIntegratorStateEstimator(
    ObstacleStateEstimator[
        NumPyIntegratorObstacleStatesHistory,
        NumPyIntegratorObstacleStates,
        NumPyIntegratorObstacleInputs,
        NumPyIntegratorObstacleCovariances,
    ]
):
    """Kalman Filter state estimator for integrator model obstacles."""

    process_noise_covariance: Float[Array, "D_x D_x"]
    observation_noise_covariance: Float[Array, "D_o D_o"]
    model: NumPyIntegratorStateEstimationModel
    estimator: NumPyKalmanFilter

    @staticmethod
    def create(
        *,
        time_step_size: float,
        process_noise_covariance: NumPyNoiseCovarianceDescription,
        observation_noise_covariance: NumPyNoiseCovarianceArrayDescription,
        initial_state_covariance: Float[Array, "D_x D_x"] | None = None,
        observation_dimension: int | None = None,
        noise_model: NumPyNoiseModelProvider | None = None,
    ) -> "NumPyKfIntegratorStateEstimator":
        """Creates an integrator state estimator based on the Kalman Filter with the
        specified noise covariances.

        Args:
            time_step_size: The time step size for the integrator.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a single scalar representing
                isotropic noise across all state dimensions.
            observation_noise_covariance: The observation noise covariance, either as
                a full matrix, a vector of diagonal entries, or a single scalar
                representing isotropic noise across all state dimensions.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for unobserved velocities.
            observation_dimension: The observation dimension for the Kalman filter.
                Mandatory if noise/state covariances are not specified/specified as scalars.
        """
        observation_dimension = observation_dimension_from(
            process_noise_covariance=process_noise_covariance,
            observation_noise_covariance=observation_noise_covariance,
            initial_state_covariance=initial_state_covariance,
            observation_dimension=observation_dimension,
        )

        return NumPyKfIntegratorStateEstimator(
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance,
                dimension=kf_state_dimension_for(observation_dimension),
            ),
            observation_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=observation_dimension
            ),
            model=NumPyIntegratorStateEstimationModel.create(
                time_step_size=time_step_size,
                observation_dimension=observation_dimension,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=NumPyKalmanFilter.create(noise_model=noise_model),
        )

    def estimate_from(
        self, history: NumPyIntegratorObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        NumPyIntegratorObstacleStates,
        NumPyIntegratorObstacleInputs,
        NumPyIntegratorObstacleCovariances,
    ]:
        """Estimate states and velocities using Kalman filtering."""
        assert history.horizon > 0, (
            "History must contain at least one state for estimation."
        )
        assert history.dimension == self.model.observation_dimension, (
            f"History dimension {history.dimension} does not match expected "
            f"observation dimension {self.model.observation_dimension}."
        )

        estimate = self.estimator.filter(
            observations=history.array,
            initial_state_covariance=self.model.initial_state_covariance,
            state_transition_matrix=self.model.state_transition_matrix,
            process_noise_covariance=self.process_noise_covariance,
            observation_noise_covariance=self.observation_noise_covariance,
            observation_matrix=self.model.observation_matrix,
        )

        return EstimatedObstacleStates(
            states=self.model.states_from(estimate),
            inputs=self.model.inputs_from(estimate),
            covariance=estimate.covariance,
        )


def validate_periodic_state_limits(state_limits: tuple[float, float] | None) -> None:
    assert state_limits is not None, (
        "Periodic boundaries require explicit state limits."
    )

    lower, upper = state_limits
    assert np.isfinite(lower) and np.isfinite(upper), (
        "Periodic boundaries must be finite."
    )

    assert upper > lower, (
        "Periodic boundaries require upper limit to be greater than lower limit."
    )


def simulate_with_state_limits(
    *,
    inputs: Float[Array, "T D_x M"],
    initial_state: Float[Array, " D_x"],
    time_step: float,
    state_limits: tuple[float, float],
) -> Float[Array, "T D_x M"]:
    deltas = inputs * time_step

    states = np.empty_like(deltas)
    states[0] = np.clip(initial_state[:, np.newaxis] + deltas[0], *state_limits)

    for t in range(1, deltas.shape[0]):
        states[t] = np.clip(states[t - 1] + deltas[t], *state_limits)

    return states


def wrap_periodic_state(
    *,
    state: Float[Array, " D_x"],
    state_limits: tuple[float, float],
) -> Float[Array, " D_x"]:
    lower, upper = state_limits
    period = upper - lower
    wrapped = lower + np.mod(state - lower, period)
    return np.where(state == upper, upper, wrapped)


def wrap_periodic_batch(
    *,
    states: Float[Array, "T D_x M"],
    state_limits: tuple[float, float],
) -> Float[Array, "T D_x M"]:
    lower, upper = state_limits
    period = upper - lower
    wrapped = lower + np.mod(states - lower, period)
    return np.where(states == upper, upper, wrapped)


def simulate(
    *,
    inputs: Float[Array, "T D_x N"],
    initial_states: Float[Array, "D_x N"],
    time_step: float,
) -> Float[Array, "T D_x N"]:
    states = initial_states + np.cumsum(inputs * time_step, axis=0)
    return states



from .bicycle import (
    BICYCLE_D_X as BICYCLE_D_X,
    BICYCLE_D_U as BICYCLE_D_U,
    BICYCLE_D_O as BICYCLE_D_O,
    BicycleD_x as BicycleD_x,
    BicycleD_u as BicycleD_u,
    BicycleD_o as BicycleD_o,
    BicycleState as BicycleState,
    BicycleStateSequence as BicycleStateSequence,
    BicycleStateBatch as BicycleStateBatch,
    BicyclePositions as BicyclePositions,
    BicycleControlInputSequence as BicycleControlInputSequence,
    BicycleControlInputBatch as BicycleControlInputBatch,
    NumPyBicycleObstacleStatesHistory as NumPyBicycleObstacleStatesHistory,
    JaxBicycleObstacleStatesHistory as JaxBicycleObstacleStatesHistory,
)
from .integrator import (
    IntegratorState as IntegratorState,
    IntegratorStateSequence as IntegratorStateSequence,
    IntegratorStateBatch as IntegratorStateBatch,
    IntegratorControlInputSequence as IntegratorControlInputSequence,
    IntegratorControlInputBatch as IntegratorControlInputBatch,
    NumPyIntegratorState as NumPyIntegratorState,
    NumPyIntegratorStateSequence as NumPyIntegratorStateSequence,
    NumPyIntegratorStateBatch as NumPyIntegratorStateBatch,
    NumPyIntegratorControlInputSequence as NumPyIntegratorControlInputSequence,
    NumPyIntegratorControlInputBatch as NumPyIntegratorControlInputBatch,
    NumPyIntegratorObstacleStatesHistory as NumPyIntegratorObstacleStatesHistory,
    JaxIntegratorState as JaxIntegratorState,
    JaxIntegratorStateSequence as JaxIntegratorStateSequence,
    JaxIntegratorStateBatch as JaxIntegratorStateBatch,
    JaxIntegratorControlInputSequence as JaxIntegratorControlInputSequence,
    JaxIntegratorControlInputBatch as JaxIntegratorControlInputBatch,
    JaxIntegratorObstacleStatesHistory as JaxIntegratorObstacleStatesHistory,
)
from .unicycle import (
    UNICYCLE_D_X as UNICYCLE_D_X,
    UNICYCLE_D_U as UNICYCLE_D_U,
    UNICYCLE_D_O as UNICYCLE_D_O,
    UnicycleD_x as UnicycleD_x,
    UnicycleD_u as UnicycleD_u,
    UnicycleD_o as UnicycleD_o,
    UnicycleState as UnicycleState,
    UnicycleStateSequence as UnicycleStateSequence,
    UnicycleStateBatch as UnicycleStateBatch,
    UnicyclePositions as UnicyclePositions,
    UnicycleControlInputSequence as UnicycleControlInputSequence,
    UnicycleControlInputBatch as UnicycleControlInputBatch,
    NumPyUnicycleObstacleStatesHistory as NumPyUnicycleObstacleStatesHistory,
    JaxUnicycleObstacleStatesHistory as JaxUnicycleObstacleStatesHistory,
)



from typing import cast, Self, Sequence, Final, NamedTuple
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    JaxState,
    JaxStateSequence,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxUnicycleObstacleStatesHistory,
    UnicycleState,
    UnicycleStateSequence,
    UnicycleStateBatch,
    UnicycleControlInputSequence,
    UnicycleControlInputBatch,
    UnicyclePositions,
    UnicycleD_x,
    UNICYCLE_D_X,
    UnicycleD_u,
    UNICYCLE_D_U,
    UnicycleD_o,
    UNICYCLE_D_O,
    POSE_D_O,
    DynamicalModel,
    ObstacleModel,
    ObstacleStateEstimator,
    EstimatedObstacleStates,
    JaxGaussianBelief,
    JaxNoiseCovarianceDescription,
    JaxNoiseModelProvider,
)
from faran.filters import (
    JaxExtendedKalmanFilter,
    JaxUnscentedKalmanFilter,
    jax_kalman_filter,
)
from faran.obstacles import JaxObstacle2dPoses
from faran.models.common import SMALL_UNCERTAINTY, LARGE_UNCERTAINTY
from faran.models.accelerated import invalid_obstacle_filter_from

from jaxtyping import Array as JaxArray, Float, Scalar
from numtypes import D

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np

NO_LIMITS: Final = (jnp.asarray(-jnp.inf), jnp.asarray(jnp.inf))

UNICYCLE_ESTIMATION_D_X: Final = 5
UNICYCLE_OBSERVATION_D_O: Final = 3

type UnicycleEstimationD_x = D[5]
type UnicycleObservationD_o = D[3]

type StateArray = Float[JaxArray, f"{UNICYCLE_D_X}"]
type ControlInputSequenceArray = Float[JaxArray, f"T {UNICYCLE_D_U}"]
type StateBatchArray = Float[JaxArray, f"T {UNICYCLE_D_X} M"]
type ControlInputBatchArray = Float[JaxArray, f"T {UNICYCLE_D_U} M"]

type StatesAtTimeStep = Float[JaxArray, f"{UNICYCLE_D_X} M"]
type ControlInputsAtTimeStep = Float[JaxArray, f"{UNICYCLE_D_U} M"]

type EstimationStateCovarianceArray = Float[
    JaxArray, f"{UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X}"
]
type ProcessNoiseCovarianceArray = Float[
    JaxArray, f"{UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X}"
]
type ObservationNoiseCovarianceArray = Float[
    JaxArray, f"{UNICYCLE_OBSERVATION_D_O} {UNICYCLE_OBSERVATION_D_O}"
]
type ObservationMatrix = Float[
    JaxArray, f"{UNICYCLE_OBSERVATION_D_O} {UNICYCLE_ESTIMATION_D_X}"
]

type UnicycleGaussianBelief = JaxGaussianBelief
type JaxUnicycleObstacleCovariances = Float[
    JaxArray, f"{UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X} K"
]
type KalmanFilter = JaxExtendedKalmanFilter | JaxUnscentedKalmanFilter


@jaxtyped
@dataclass(frozen=True)
class JaxUnicycleState(UnicycleState, JaxState):
    """Kinematic unicycle state: [x, y, heading]."""

    _array: StateArray

    @staticmethod
    def create(
        *,
        x: float | Scalar,
        y: float | Scalar,
        heading: float | Scalar,
    ) -> "JaxUnicycleState":
        return JaxUnicycleState(jnp.array([x, y, heading]))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, " UnicycleD_x"]:
        return np.asarray(self.array)

    @property
    def dimension(self) -> UnicycleD_x:
        return cast(UnicycleD_x, self.array.shape[0])

    @property
    def x(self) -> float:
        return float(self.array[0])

    @property
    def y(self) -> float:
        return float(self.array[1])

    @property
    def heading(self) -> float:
        return float(self.array[2])

    @property
    def array(self) -> StateArray:
        return self._array

    @property
    def x_scalar(self) -> Scalar:
        return self.array[0]

    @property
    def y_scalar(self) -> Scalar:
        return self.array[1]

    @property
    def heading_scalar(self) -> Scalar:
        return self.array[2]


@dataclass(kw_only=True, frozen=True)
class JaxUnicycleStateSequence(UnicycleStateSequence, JaxStateSequence):
    batch: "JaxUnicycleStateBatch"
    rollout: int

    @staticmethod
    def of_states(states: Sequence[JaxUnicycleState]) -> "JaxUnicycleStateSequence":
        assert len(states) > 0, "States sequence must not be empty."

        array = jnp.stack([state.array for state in states], axis=0)[:, :, jnp.newaxis]

        return JaxUnicycleStateSequence(
            batch=JaxUnicycleStateBatch.wrap(array), rollout=0
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T UnicycleD_x"]:
        return np.asarray(self.array)

    def step(self, index: int) -> JaxUnicycleState:
        return JaxUnicycleState(self.array[index])

    def batched(self) -> "JaxUnicycleStateBatch":
        return JaxUnicycleStateBatch.wrap(self.array[..., jnp.newaxis])

    def x(self) -> Float[Array, " T"]:
        return np.asarray(self.array[:, 0])

    def y(self) -> Float[Array, " T"]:
        return np.asarray(self.array[:, 1])

    def heading(self) -> Float[Array, " T"]:
        return np.asarray(self.array[:, 2])

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_x:
        return cast(UnicycleD_x, self.array.shape[1])

    @property
    def array(self) -> Float[JaxArray, f"T {UNICYCLE_D_X}"]:
        return self.batch.array[:, :, self.rollout]


@jaxtyped
@dataclass(frozen=True)
class JaxUnicycleStateBatch(UnicycleStateBatch, JaxStateBatch):
    _array: StateBatchArray

    @staticmethod
    def wrap(
        array: StateBatchArray,
    ) -> "JaxUnicycleStateBatch":
        return JaxUnicycleStateBatch(array)

    @staticmethod
    def of_states(states: Sequence[JaxUnicycleState]) -> "JaxUnicycleStateBatch":
        assert len(states) > 0, "States sequence must not be empty."

        array = jnp.stack([state.array for state in states], axis=0)[:, :, jnp.newaxis]

        return JaxUnicycleStateBatch(array)

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T UnicycleD_x M"]:
        return np.asarray(self.array)

    def heading(self) -> Float[Array, "T M"]:
        return np.asarray(self.array[:, 2, :])

    def rollout(self, index: int) -> JaxUnicycleStateSequence:
        return JaxUnicycleStateSequence(batch=self, rollout=index)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_x:
        return cast(UnicycleD_x, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def positions(self) -> "JaxUnicyclePositions":
        return JaxUnicyclePositions(self)

    @property
    def array(self) -> StateBatchArray:
        return self._array

    @property
    def heading_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 2, :]


@jaxtyped
@dataclass(frozen=True)
class JaxUnicyclePositions(UnicyclePositions):
    batch: JaxUnicycleStateBatch

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 M"]:
        return self._numpy_array

    def x(self) -> Float[Array, "T M"]:
        return self._numpy_array[:, 0, :]

    def y(self) -> Float[Array, "T M"]:
        return self._numpy_array[:, 1, :]

    @property
    def x_array(self) -> Float[JaxArray, "T M"]:
        return self.batch.array[:, 0, :]

    @property
    def y_array(self) -> Float[JaxArray, "T M"]:
        return self.batch.array[:, 1, :]

    @property
    def array(self) -> Float[JaxArray, "T 2 M"]:
        return self.batch.array[:, :2, :]

    @property
    def horizon(self) -> int:
        return self.batch.horizon

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def rollout_count(self) -> int:
        return self.batch.rollout_count

    @cached_property
    def _numpy_array(self) -> Float[Array, "T 2 M"]:
        return np.asarray(self.batch.array[:, :2, :])


@jaxtyped
@dataclass(frozen=True)
class JaxUnicycleControlInputSequence(
    UnicycleControlInputSequence, JaxControlInputSequence
):
    """Control inputs: [linear velocity, angular velocity]."""

    _array: ControlInputSequenceArray

    @staticmethod
    def zeroes(horizon: int) -> "JaxUnicycleControlInputSequence":
        return JaxUnicycleControlInputSequence(jnp.zeros((horizon, UNICYCLE_D_U)))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T UnicycleD_u"]:
        return self._numpy_array

    def similar(self, *, array: Float[JaxArray, "L D_u"]) -> Self:

        return self.__class__(array)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_u:
        return cast(UnicycleD_u, self.array.shape[1])

    @property
    def array(self) -> ControlInputSequenceArray:
        return self._array

    @property
    def linear_velocity_array(self) -> Float[JaxArray, " T"]:
        return self.array[:, 0]

    @property
    def angular_velocity_array(self) -> Float[JaxArray, " T"]:
        return self.array[:, 1]

    @cached_property
    def _numpy_array(self) -> Float[Array, "T UnicycleD_u"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxUnicycleControlInputBatch(UnicycleControlInputBatch, JaxControlInputBatch):
    _array: ControlInputBatchArray

    @staticmethod
    def zero(*, horizon: int, rollout_count: int = 1) -> "JaxUnicycleControlInputBatch":
        array = jnp.zeros((horizon, UNICYCLE_D_U, rollout_count))

        return JaxUnicycleControlInputBatch(array)

    @staticmethod
    def create(
        *, array: Float[Array, "T UnicycleD_u M"] | ControlInputBatchArray
    ) -> "JaxUnicycleControlInputBatch":
        return JaxUnicycleControlInputBatch(jnp.asarray(array))

    @staticmethod
    def of(sequence: JaxUnicycleControlInputSequence) -> "JaxUnicycleControlInputBatch":

        return JaxUnicycleControlInputBatch(sequence.array[..., jnp.newaxis])

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T UnicycleD_u M"]:
        return self._numpy_array

    def linear_velocity(self) -> Float[Array, "T M"]:
        return self._numpy_array[:, 0, :]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_u:
        return cast(UnicycleD_u, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> ControlInputBatchArray:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T UnicycleD_u M"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxUnicycleObstacleStates:
    _array: Float[JaxArray, f"{UNICYCLE_D_O} K"]

    @staticmethod
    def wrap(
        array: Float[Array, "UnicycleD_o K"] | Float[JaxArray, f"{UNICYCLE_D_O} K"],
    ) -> "JaxUnicycleObstacleStates":
        return JaxUnicycleObstacleStates(jnp.asarray(array))

    @staticmethod
    def create(
        *,
        x: Float[Array, " K"] | Float[JaxArray, " K"],
        y: Float[Array, " K"] | Float[JaxArray, " K"],
        heading: Float[Array, " K"] | Float[JaxArray, " K"],
    ) -> "JaxUnicycleObstacleStates":
        return JaxUnicycleObstacleStates(jnp.stack([x, y, heading], axis=0))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "UnicycleD_o K"]:
        return self._numpy_array

    def x(self) -> Float[Array, " K"]:
        return self._numpy_array[0, :]

    def y(self) -> Float[Array, " K"]:
        return self._numpy_array[1, :]

    def heading(self) -> Float[Array, " K"]:
        return self._numpy_array[2, :]

    @property
    def dimension(self) -> UnicycleD_o:
        return cast(UnicycleD_o, self.array.shape[0])

    @property
    def count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[JaxArray, f"{UNICYCLE_D_O} K"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "UnicycleD_o K"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxUnicycleObstacleStateSequences:
    _array: Float[JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} K"]
    _covariance: Float[
        JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X} K"
    ]

    @staticmethod
    def wrap(
        *,
        array: Float[JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} K"],
        covariance: Float[
            JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X} K"
        ],
    ) -> "JaxUnicycleObstacleStateSequences":
        return JaxUnicycleObstacleStateSequences(array, covariance)

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T UnicycleEstimationD_x K"]:
        return self._numpy_array

    def x(self) -> Float[Array, "T K"]:
        return self._numpy_array[:, 0, :]

    def y(self) -> Float[Array, "T K"]:
        return self._numpy_array[:, 1, :]

    def heading(self) -> Float[Array, "T K"]:
        return self._numpy_array[:, 2, :]

    def covariance(
        self,
    ) -> Float[JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X} K"]:
        return self._covariance

    def pose(self) -> JaxObstacle2dPoses:
        return JaxObstacle2dPoses.create(
            x=self.x_array,
            y=self.y_array,
            heading=self.heading_array,
            covariance=self.pose_covariance_array,
        )

    @property
    def horizon(self) -> int:
        return self._array.shape[0]

    @property
    def dimension(self) -> UnicycleEstimationD_x:
        return cast(UnicycleEstimationD_x, self._array.shape[1])

    @property
    def count(self) -> int:
        return self._array.shape[2]

    @property
    def array(self) -> Float[JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} K"]:
        return self._array

    @property
    def x_array(self) -> Float[JaxArray, "T K"]:
        return self._array[:, 0, :]

    @property
    def y_array(self) -> Float[JaxArray, "T K"]:
        return self._array[:, 1, :]

    @property
    def heading_array(self) -> Float[JaxArray, "T K"]:
        return self._array[:, 2, :]

    @property
    def covariance_array(
        self,
    ) -> Float[JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X} K"]:
        return self._covariance

    @property
    def pose_covariance_array(self) -> Float[JaxArray, f"T {POSE_D_O} {POSE_D_O} K"]:
        return self.covariance_array[:, :3, :3, :]

    @cached_property
    def _numpy_array(self) -> Float[Array, "T UnicycleEstimationD_x K"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxUnicycleObstacleInputs:
    _linear_velocities: Float[JaxArray, " K"]
    _angular_velocities: Float[JaxArray, " K"]

    @staticmethod
    def wrap(
        array: Float[JaxArray, f"{UNICYCLE_D_U} K"] | Float[Array, "UnicycleD_u K"],
    ) -> "JaxUnicycleObstacleInputs":
        linear, angular = jnp.asarray(array)
        return JaxUnicycleObstacleInputs(
            _linear_velocities=linear, _angular_velocities=angular
        )

    @staticmethod
    def create(
        *,
        linear_velocities: Float[JaxArray, " K"] | Float[Array, " K"],
        angular_velocities: Float[JaxArray, " K"] | Float[Array, " K"],
    ) -> "JaxUnicycleObstacleInputs":
        return JaxUnicycleObstacleInputs(
            _linear_velocities=jnp.asarray(linear_velocities),
            _angular_velocities=jnp.asarray(angular_velocities),
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "UnicycleD_u K"]:
        return self._numpy_array

    def linear_velocities(self) -> Float[Array, " K"]:
        return self._numpy_array[0, :]

    def angular_velocities(self) -> Float[Array, " K"]:
        return self._numpy_array[1, :]

    def zeroed(
        self, *, linear_velocity: bool = False, angular_velocity: bool = False
    ) -> "JaxUnicycleObstacleInputs":
        """Returns a version of the inputs with the specified components zeroed out."""
        return JaxUnicycleObstacleInputs(
            _linear_velocities=jnp.zeros_like(self.linear_velocities())
            if linear_velocity
            else self._linear_velocities,
            _angular_velocities=jnp.zeros_like(self.angular_velocities())
            if angular_velocity
            else self._angular_velocities,
        )

    @property
    def dimension(self) -> UnicycleD_u:
        return UNICYCLE_D_U

    @property
    def count(self) -> int:
        return self._linear_velocities.shape[0]

    @property
    def linear_velocities_array(self) -> Float[JaxArray, " K"]:
        return self._linear_velocities

    @property
    def angular_velocities_array(self) -> Float[JaxArray, " K"]:
        return self._angular_velocities

    @property
    def array(self) -> Float[JaxArray, f"{UNICYCLE_D_U} K"]:
        return jnp.stack([self._linear_velocities, self._angular_velocities], axis=0)

    @cached_property
    def _numpy_array(self) -> Float[Array, "UnicycleD_u K"]:
        return np.asarray(self.array)


@dataclass(kw_only=True, frozen=True)
class JaxUnicycleModel(
    DynamicalModel[
        JaxUnicycleState,
        JaxUnicycleStateSequence,
        JaxUnicycleStateBatch,
        JaxUnicycleControlInputSequence,
        JaxUnicycleControlInputBatch,
    ],
):
    """Kinematic unicycle with direct velocity control, Euler-integrated with configurable limits."""

    _time_step_size: float
    time_step_size_scalar: Scalar
    speed_limits: tuple[Scalar, Scalar]
    angular_velocity_limits: tuple[Scalar, Scalar]

    @staticmethod
    def create(
        *,
        time_step_size: float,
        speed_limits: tuple[float, float] | None = None,
        angular_velocity_limits: tuple[float, float] | None = None,
    ) -> "JaxUnicycleModel":

        return JaxUnicycleModel(
            _time_step_size=time_step_size,
            time_step_size_scalar=jnp.asarray(time_step_size),
            speed_limits=wrap(speed_limits) if speed_limits is not None else NO_LIMITS,
            angular_velocity_limits=wrap(angular_velocity_limits)
            if angular_velocity_limits is not None
            else NO_LIMITS,
        )

    def simulate(
        self,
        inputs: JaxUnicycleControlInputBatch,
        initial_state: JaxUnicycleState,
    ) -> JaxUnicycleStateBatch:
        rollout_count = inputs.rollout_count

        initial = jnp.stack(
            [
                jnp.full(rollout_count, initial_state.x_scalar),
                jnp.full(rollout_count, initial_state.y_scalar),
                jnp.full(rollout_count, initial_state.heading_scalar),
            ]
        )

        return JaxUnicycleStateBatch(
            simulate(
                inputs.array,
                initial,
                time_step_size=self.time_step_size_scalar,
                speed_limits=self.speed_limits,
                angular_velocity_limits=self.angular_velocity_limits,
            )
        )

    def step(
        self, inputs: JaxUnicycleControlInputSequence, state: JaxUnicycleState
    ) -> JaxUnicycleState:
        return JaxUnicycleState(
            step(
                state.array.reshape(-1, 1),
                inputs.array[0].reshape(-1, 1),
                time_step_size=self.time_step_size_scalar,
                speed_limits=self.speed_limits,
                angular_velocity_limits=self.angular_velocity_limits,
            )[:, 0]
        )

    def forward(
        self, inputs: JaxUnicycleControlInputSequence, state: JaxUnicycleState
    ) -> JaxUnicycleStateSequence:
        return self.simulate(JaxUnicycleControlInputBatch.of(inputs), state).rollout(0)

    @property
    def time_step_size(self) -> float:
        return self._time_step_size


class JaxUnicycleStateEstimationModel(NamedTuple):
    """Kinematic unicycle model used for state estimation."""

    time_step_size: Scalar
    initial_state_covariance: EstimationStateCovarianceArray

    @staticmethod
    def create(
        *,
        time_step_size: float,
        initial_state_covariance: Float[
            Array, "UnicycleEstimationD_x UnicycleEstimationD_x"
        ]
        | EstimationStateCovarianceArray
        | None = None,
    ) -> "JaxUnicycleStateEstimationModel":
        if initial_state_covariance is None:
            initial_state_covariance = (
                JaxUnicycleStateEstimationModel.default_initial_state_covariance()
            )

        return JaxUnicycleStateEstimationModel(
            time_step_size=jnp.asarray(time_step_size),
            initial_state_covariance=jnp.asarray(initial_state_covariance),
        )

    @staticmethod
    def default_initial_state_covariance() -> EstimationStateCovarianceArray:
        # NOTE: Sure of pose, unsure of velocities.
        return jnp.diag(
            jnp.array(
                [
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                ]
            )
        )

    @jax.jit
    @jaxtyped
    def __call__(self, state: Float[JaxArray, "D_x K"]) -> Float[JaxArray, "D_x K"]:
        return jax.vmap(self.step, in_axes=1, out_axes=1)(state)

    @jax.jit
    @jaxtyped
    def jacobian(self, state: Float[JaxArray, "D_x K"]) -> Float[JaxArray, "D_x D_x K"]:
        return jax.vmap(jax.jacfwd(self.step))(state.T).transpose(1, 2, 0)

    @jax.jit
    @jaxtyped
    def step(self, state: Float[JaxArray, " D_x"]) -> Float[JaxArray, " D_x"]:
        dt = self.time_step_size
        x, y, theta, v, omega = state
        return jnp.array(
            [
                x + v * jnp.cos(theta) * dt,
                y + v * jnp.sin(theta) * dt,
                theta + omega * dt,
                v,
                omega,
            ]
        )

    def observations_from(
        self, history: JaxUnicycleObstacleStatesHistory
    ) -> Float[JaxArray, "T D_z K"]:
        return jnp.stack(
            [history.x_array, history.y_array, history.heading_array], axis=1
        )

    def states_from(self, belief: UnicycleGaussianBelief) -> JaxUnicycleObstacleStates:
        return JaxUnicycleObstacleStates.wrap(belief.mean[:3, :])

    def inputs_from(self, belief: UnicycleGaussianBelief) -> JaxUnicycleObstacleInputs:
        return cast(
            JaxUnicycleObstacleInputs,
            JaxUnicycleObstacleInputs.wrap(belief.mean[3:, :]),
        )

    def initial_belief_from(
        self,
        *,
        states: JaxUnicycleObstacleStates,
        inputs: JaxUnicycleObstacleInputs,
        covariances: JaxUnicycleObstacleCovariances | None = None,
    ) -> UnicycleGaussianBelief:
        augmented = jnp.concatenate([states.array, inputs.array], axis=0)

        if covariances is None:
            # NOTE: No covariance means we are "certain" about the states.
            covariances = jnp.broadcast_to(
                jnp.eye(UNICYCLE_ESTIMATION_D_X)[:, :, jnp.newaxis] * SMALL_UNCERTAINTY,
                (UNICYCLE_ESTIMATION_D_X, UNICYCLE_ESTIMATION_D_X, states.count),
            )

        return JaxGaussianBelief(mean=augmented, covariance=covariances)

    @property
    def observation_matrix(self) -> ObservationMatrix:
        # NOTE: We observe the pose directly.
        return jnp.array(
            [
                [1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
            ]
        )


@dataclass(kw_only=True, frozen=True)
class JaxUnicycleObstacleModel(
    ObstacleModel[
        JaxUnicycleObstacleStatesHistory,
        JaxUnicycleObstacleStates,
        JaxUnicycleObstacleInputs,
        JaxUnicycleObstacleCovariances,
        JaxUnicycleObstacleStateSequences,
    ]
):
    """Propagates unicycle kinematics forward given states and control inputs."""

    model: "JaxUnicycleStateEstimationModel"
    process_noise_covariance: ProcessNoiseCovarianceArray
    predictor: KalmanFilter

    @staticmethod
    def unscented(
        *,
        time_step_size: float,
        process_noise_covariance: JaxNoiseCovarianceDescription = 1e-3,
        sigma_point_spread: float = 1.0,
        prior_knowledge: float = 2.0,
    ) -> "JaxUnicycleObstacleModel":
        """Creates a unicycle obstacle model for obstacle state prediction. This
        model uses the unscented transform for propagating state information through
        the nonlinear unicycle dynamics.

        Args:
            time_step_size: Time step size for state propagation.
            process_noise_covariance: The process noise covariance, either as a
                full covariance array, a diagonal covariance vector, or a scalar
                variance representing isotropic noise.
            sigma_point_spread: Spread of sigma points (α) for the unscented
                transform, controlling how far the sigma points are from the mean.
            prior_knowledge: Prior knowledge (β) about the state distribution
                for the unscented transform.
        """
        standardized_covariance = jax_kalman_filter.standardize_noise_covariance(
            process_noise_covariance, dimension=UNICYCLE_ESTIMATION_D_X
        )
        return JaxUnicycleObstacleModel(
            model=JaxUnicycleStateEstimationModel.create(
                time_step_size=time_step_size,
            ),
            process_noise_covariance=standardized_covariance,
            predictor=JaxUnscentedKalmanFilter.create(
                alpha=sigma_point_spread, beta=prior_knowledge
            ),
        )

    def forward(
        self,
        *,
        states: JaxUnicycleObstacleStates,
        inputs: JaxUnicycleObstacleInputs,
        covariances: JaxUnicycleObstacleCovariances | None,
        horizon: int,
    ) -> JaxUnicycleObstacleStateSequences:
        means, covariance = forward(
            initial=self.model.initial_belief_from(
                states=states, inputs=inputs, covariances=covariances
            ),
            model=self.model,
            predictor=self.predictor,
            process_noise_covariance=self.process_noise_covariance,
            horizon=horizon,
        )

        return JaxUnicycleObstacleStateSequences.wrap(
            array=means, covariance=covariance
        )


@dataclass(frozen=True)
class JaxFiniteDifferenceUnicycleStateEstimator(
    ObstacleStateEstimator[
        JaxUnicycleObstacleStatesHistory,
        JaxUnicycleObstacleStates,
        JaxUnicycleObstacleInputs,
    ]
):
    time_step_size: Scalar

    @staticmethod
    def create(*, time_step_size: float) -> "JaxFiniteDifferenceUnicycleStateEstimator":
        return JaxFiniteDifferenceUnicycleStateEstimator(
            time_step_size=jnp.asarray(time_step_size)
        )

    def estimate_from(
        self, history: JaxUnicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        JaxUnicycleObstacleStates, JaxUnicycleObstacleInputs, None
    ]:
        """Estimates current states and inputs from position/heading history using finite differences.

        Computes the following quantities from the unicycle model (requires T ≥ 2, otherwise returns zeros):

        **Linear velocity**:
            Projection of displacement onto the heading direction (negative for reverse):
            $$v_t = \\frac{(x_t - x_{t-1}) \\cos(\\theta_t) + (y_t - y_{t-1}) \\sin(\\theta_t)}{\\Delta t}$$

        **Angular velocity**:
            Change in heading over time:
            $$\\omega_t = \\frac{\\theta_t - \\theta_{t-1}}{\\Delta t}$$

        Args:
            history: History of observed poses with at least one entry.
        """
        assert history.horizon > 0, "History must contain at least one state."

        estimated = estimate_states(
            x_history=history.x_array,
            y_history=history.y_array,
            heading_history=history.heading_array,
            time_step_size=self.time_step_size,
        )

        return EstimatedObstacleStates(
            states=JaxUnicycleObstacleStates.create(
                x=estimated.x,
                y=estimated.y,
                heading=estimated.heading,
            ),
            inputs=JaxUnicycleObstacleInputs.create(
                linear_velocities=estimated.linear_velocities,
                angular_velocities=estimated.angular_velocities,
            ),
            covariance=None,
        )


@dataclass(frozen=True)
class JaxKfUnicycleStateEstimator(
    ObstacleStateEstimator[
        JaxUnicycleObstacleStatesHistory,
        JaxUnicycleObstacleStates,
        JaxUnicycleObstacleInputs,
        JaxUnicycleObstacleCovariances,
    ]
):
    """Kalman Filter state estimator for unicycle model obstacles."""

    process_noise_covariance: ProcessNoiseCovarianceArray
    observation_noise_covariance: ObservationNoiseCovarianceArray
    model: JaxUnicycleStateEstimationModel
    estimator: KalmanFilter

    @staticmethod
    def ekf(
        *,
        time_step_size: float,
        process_noise_covariance: JaxNoiseCovarianceDescription,
        observation_noise_covariance: JaxNoiseCovarianceDescription,
        initial_state_covariance: Float[
            Array, "UnicycleEstimationD_x UnicycleEstimationD_x"
        ]
        | EstimationStateCovarianceArray
        | None = None,
        noise_model: JaxNoiseModelProvider | None = None,
    ) -> "JaxKfUnicycleStateEstimator":
        """Creates an EKF state estimator for the unicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for unobserved velocities.
        """
        return JaxKfUnicycleStateEstimator(
            process_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=UNICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=UNICYCLE_OBSERVATION_D_O
            ),
            model=JaxUnicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=JaxExtendedKalmanFilter.create(noise_model=noise_model),
        )

    @staticmethod
    def ukf(
        *,
        time_step_size: float,
        process_noise_covariance: JaxNoiseCovarianceDescription,
        observation_noise_covariance: JaxNoiseCovarianceDescription,
        initial_state_covariance: Float[
            Array, "UnicycleEstimationD_x UnicycleEstimationD_x"
        ]
        | EstimationStateCovarianceArray
        | None = None,
        noise_model: JaxNoiseModelProvider | None = None,
    ) -> "JaxKfUnicycleStateEstimator":
        """Creates a UKF state estimator for the unicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for unobserved velocities.
        """
        return JaxKfUnicycleStateEstimator(
            process_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=UNICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=jax_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=UNICYCLE_OBSERVATION_D_O
            ),
            model=JaxUnicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=JaxUnscentedKalmanFilter.create(noise_model=noise_model),
        )

    def estimate_from(
        self, history: JaxUnicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        JaxUnicycleObstacleStates,
        JaxUnicycleObstacleInputs,
        JaxUnicycleObstacleCovariances,
    ]:
        estimate = self.estimator.filter(
            self.model.observations_from(history),
            initial_state_covariance=self.model.initial_state_covariance,
            state_transition=self.model,
            process_noise_covariance=self.process_noise_covariance,
            observation_noise_covariance=self.observation_noise_covariance,
            observation_matrix=self.model.observation_matrix,
        )

        return EstimatedObstacleStates(
            states=self.model.states_from(estimate),
            inputs=self.model.inputs_from(estimate),
            covariance=estimate.covariance,
        )


class EstimatedUnicycleObstacleStates(NamedTuple):
    x: Float[JaxArray, " K"]
    y: Float[JaxArray, " K"]
    heading: Float[JaxArray, " K"]
    linear_velocities: Float[JaxArray, " K"]
    angular_velocities: Float[JaxArray, " K"]


def wrap(limits: tuple[float, float]) -> tuple[Scalar, Scalar]:
    return (jnp.asarray(limits[0]), jnp.asarray(limits[1]))


@jax.jit
@jaxtyped
def simulate(
    inputs: ControlInputBatchArray,
    initial: StatesAtTimeStep,
    *,
    time_step_size: Scalar,
    speed_limits: tuple[Scalar, Scalar],
    angular_velocity_limits: tuple[Scalar, Scalar],
) -> StateBatchArray:
    def do_step(
        state: StatesAtTimeStep, control: ControlInputsAtTimeStep
    ) -> tuple[StatesAtTimeStep, StatesAtTimeStep]:
        new_state = step(
            state,
            control,
            time_step_size=time_step_size,
            speed_limits=speed_limits,
            angular_velocity_limits=angular_velocity_limits,
        )
        return new_state, new_state

    _, states = jax.lax.scan(do_step, initial, inputs)
    return states


@jax.jit
@jaxtyped
def step(
    state: StatesAtTimeStep,
    control: ControlInputsAtTimeStep,
    *,
    time_step_size: Scalar,
    speed_limits: tuple[Scalar, Scalar],
    angular_velocity_limits: tuple[Scalar, Scalar],
) -> StatesAtTimeStep:
    x, y, theta = state[0], state[1], state[2]
    v, omega = control[0], control[1]
    linear_velocity = jnp.clip(v, *speed_limits)
    angular_velocity = jnp.clip(omega, *angular_velocity_limits)

    new_x = x + linear_velocity * jnp.cos(theta) * time_step_size
    new_y = y + linear_velocity * jnp.sin(theta) * time_step_size
    new_theta = theta + angular_velocity * time_step_size

    return jnp.stack([new_x, new_y, new_theta])


@jax.jit
@jaxtyped
def estimate_states(
    *,
    x_history: Float[JaxArray, "T K"],
    y_history: Float[JaxArray, "T K"],
    heading_history: Float[JaxArray, "T K"],
    time_step_size: Scalar,
) -> "EstimatedUnicycleObstacleStates":
    filter_invalid = invalid_obstacle_filter_from(
        x_history, y_history, heading_history, check_recent=2
    )

    return EstimatedUnicycleObstacleStates(
        x=x_history[-1],
        y=y_history[-1],
        heading=heading_history[-1],
        linear_velocities=filter_invalid(
            estimate_linear_velocities(
                x_history=x_history,
                y_history=y_history,
                heading_history=heading_history,
                time_step_size=time_step_size,
            )
        ),
        angular_velocities=filter_invalid(
            estimate_angular_velocities(
                heading_history=heading_history,
                time_step_size=time_step_size,
            )
        ),
    )


@jax.jit
@jaxtyped
def estimate_linear_velocities(
    *,
    x_history: Float[JaxArray, "T K"],
    y_history: Float[JaxArray, "T K"],
    heading_history: Float[JaxArray, "T K"],
    time_step_size: Scalar,
) -> Float[JaxArray, " K"]:
    horizon = x_history.shape[0]
    obstacle_count = x_history.shape[1]

    def estimate(
        *,
        x_current: Float[JaxArray, " K"],
        y_current: Float[JaxArray, " K"],
        x_previous: Float[JaxArray, " K"],
        y_previous: Float[JaxArray, " K"],
        heading: Float[JaxArray, " K"],
    ) -> Float[JaxArray, " K"]:
        delta_x = x_current - x_previous
        delta_y = y_current - y_previous

        return (
            delta_x * jnp.cos(heading) + delta_y * jnp.sin(heading)
        ) / time_step_size

    if horizon < 2:
        return jnp.zeros(obstacle_count)

    return estimate(
        x_current=x_history[-1],
        y_current=y_history[-1],
        x_previous=x_history[-2],
        y_previous=y_history[-2],
        heading=heading_history[-1],
    )


@jax.jit
@jaxtyped
def estimate_angular_velocities(
    *,
    heading_history: Float[JaxArray, "T K"],
    time_step_size: Scalar,
) -> Float[JaxArray, " K"]:
    horizon = heading_history.shape[0]
    obstacle_count = heading_history.shape[1]

    def estimate(
        *,
        heading_current: Float[JaxArray, " K"],
        heading_previous: Float[JaxArray, " K"],
    ) -> Float[JaxArray, " K"]:
        return (heading_current - heading_previous) / time_step_size

    if horizon < 2:
        return jnp.zeros(obstacle_count)

    return estimate(
        heading_current=heading_history[-1], heading_previous=heading_history[-2]
    )


@eqx.filter_jit
@jaxtyped
def forward(
    *,
    initial: UnicycleGaussianBelief,
    model: JaxUnicycleStateEstimationModel,
    predictor: KalmanFilter,
    process_noise_covariance: ProcessNoiseCovarianceArray,
    horizon: int,
) -> tuple[
    Float[JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} K"],
    Float[JaxArray, f"T {UNICYCLE_ESTIMATION_D_X} {UNICYCLE_ESTIMATION_D_X} K"],
]:
    def step(
        belief: UnicycleGaussianBelief, _
    ) -> tuple[UnicycleGaussianBelief, UnicycleGaussianBelief]:
        next_belief = predictor.predict(
            belief=belief,
            state_transition=model,
            process_noise_covariance=process_noise_covariance,
        )

        return next_belief, next_belief

    _, beliefs = jax.lax.scan(step, initial, xs=None, length=horizon)

    return beliefs.mean, beliefs.covariance



from typing import Protocol
from dataclasses import dataclass

from faran.types import Array, ObstacleStatesForTimeStep, ObstacleStateObserver

from jaxtyping import Float

import numpy as np


class ObstacleStateCreator[ObstacleStatesFormTimeStepT](Protocol):
    def __call__(self, array: Float[Array, "D_o K"], /) -> ObstacleStatesFormTimeStepT:
        """Wraps the specified array of observed obstacle states in an object of the
        appropriate type."""
        ...


@dataclass(kw_only=True, frozen=True)
class NoisyObstacleStateObserver[ObstacleStatesForTimeStepT: ObstacleStatesForTimeStep]:
    inner: ObstacleStateObserver[ObstacleStatesForTimeStepT]
    to_states: ObstacleStateCreator[ObstacleStatesForTimeStepT]
    sigma: Float[Array, " D_o"]
    rng: np.random.Generator

    @staticmethod
    def decorate[OS: ObstacleStatesForTimeStep](
        observer: ObstacleStateObserver[OS],
        *,
        to_states: ObstacleStateCreator[OS],
        sigma: Float[Array, " D_o"],
        seed: int = 0,
    ) -> "NoisyObstacleStateObserver[OS]":
        """Creates an observer that adds zero-mean Gaussian noise to the observed states before delegating to
        the underlying observer.

        Args:
            observer: The underlying observer to which the noisy observations will be delegated.
            to_states: An object to create the appropriate type of obstacle states from the noisy state arrays.
            sigma: The standard deviation of the Gaussian noise to be added to each component of the observed states.
            seed: The seed for the random number generator used to generate the noise.
        """
        return NoisyObstacleStateObserver(
            inner=observer,
            to_states=to_states,
            sigma=sigma,
            rng=np.random.default_rng(seed),
        )

    def observe(self, states: ObstacleStatesForTimeStepT) -> None:
        self.inner.observe(
            self.to_states(
                (array := np.asarray(states))
                + self.rng.normal(0, self.sigma[:, np.newaxis], array.shape)
            )
        )
